# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2023 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************
import os
import sys
import tempfile
import unittest
import unittest.mock

Mock = unittest.mock.MagicMock

sys.path.append("../../")


class TestVersion(unittest.TestCase):
    def setUp(self) -> None:
        if "addonmanager_metadata" in sys.modules:
            sys.modules.pop("addonmanager_metadata")
        self.packaging_version = None
        if "packaging.version" in sys.modules:
            self.packaging_version = sys.modules["packaging.version"]
            sys.modules.pop("packaging.version")

    def tearDown(self) -> None:
        if self.packaging_version is not None:
            sys.modules["packaging.version"] = self.packaging_version

    def test_init_from_string_manual(self):
        import addonmanager_metadata as amm

        version = amm.Version()
        version._parse_string_to_tuple = unittest.mock.MagicMock()
        version._init_from_string("1.2.3beta")
        self.assertTrue(version._parse_string_to_tuple.called)

    def test_init_from_list_good(self):
        """Initialization from a list works for good input"""
        import addonmanager_metadata as amm

        test_cases = [
            {"input": (1,), "output": [1, 0, 0, ""]},
            {"input": (1, 2), "output": [1, 2, 0, ""]},
            {"input": (1, 2, 3), "output": [1, 2, 3, ""]},
            {"input": (1, 2, 3, "b1"), "output": [1, 2, 3, "b1"]},
        ]
        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                v = amm.Version(from_list=test_case["input"])
                self.assertListEqual(test_case["output"], v.version_as_list)

    def test_parse_string_to_tuple_normal(self):
        """Parsing of complete version string works for normal cases"""
        import addonmanager_metadata as amm

        cases = {
            "1": [1, 0, 0, ""],
            "1.2": [1, 2, 0, ""],
            "1.2.3": [1, 2, 3, ""],
            "1.2.3beta": [1, 2, 3, "beta"],
            "12_345.6_7.8pre-alpha": [12345, 67, 8, "pre-alpha"],
            # The above test is mostly to point out that Python gets permits underscore
            # characters in a number.
        }
        for inp, output in cases.items():
            with self.subTest(inp=inp, output=output):
                version = amm.Version()
                version._parse_string_to_tuple(inp)
                self.assertListEqual(version.version_as_list, output)

    def test_parse_string_to_tuple_invalid(self):
        """Parsing of invalid version string raises an exception"""
        import addonmanager_metadata as amm

        cases = {"One", "1,2,3", "1-2-3", "1/2/3"}
        for inp in cases:
            with self.subTest(inp=inp):
                with self.assertRaises(ValueError):
                    version = amm.Version()
                    version._parse_string_to_tuple(inp)

    def test_parse_final_entry_normal(self):
        """Parsing of the final entry works for normal cases"""
        import addonmanager_metadata as amm

        cases = {
            "3beta": (3, "beta"),
            "42.alpha": (42, ".alpha"),
            "123.45.6": (123, ".45.6"),
            "98_delta": (98, "_delta"),
            "1 and some words": (1, " and some words"),
        }
        for inp, output in cases.items():
            with self.subTest(inp=inp, output=output):
                number, text = amm.Version._parse_final_entry(inp)
                self.assertEqual(number, output[0])
                self.assertEqual(text, output[1])

    def test_parse_final_entry_invalid(self):
        """Invalid input raises an exception"""
        import addonmanager_metadata as amm

        cases = ["beta", "", ["a", "b"]]
        for case in cases:
            with self.subTest(case=case):
                with self.assertRaises(ValueError):
                    amm.Version._parse_final_entry(case)

    def test_operators_internal(self):
        """Test internal (non-package) comparison operators"""
        sys.modules["packaging.version"] = None
        import addonmanager_metadata as amm

        cases = self.given_comparison_cases()
        for case in cases:
            with self.subTest(case=case):
                first = amm.Version(case[0])
                second = amm.Version(case[1])
                self.assertEqual(first < second, case[0] < case[1])
                self.assertEqual(first > second, case[0] > case[1])
                self.assertEqual(first <= second, case[0] <= case[1])
                self.assertEqual(first >= second, case[0] >= case[1])
                self.assertEqual(first == second, case[0] == case[1])

    @staticmethod
    def given_comparison_cases():
        return [
            ("0.0.0alpha", "1.0.0alpha"),
            ("0.0.0alpha", "0.1.0alpha"),
            ("0.0.0alpha", "0.0.1alpha"),
            ("0.0.0alpha", "0.0.0beta"),
            ("0.0.0alpha", "0.0.0alpha"),
            ("1.0.0alpha", "0.0.0alpha"),
            ("0.1.0alpha", "0.0.0alpha"),
            ("0.0.1alpha", "0.0.0alpha"),
            ("0.0.0beta", "0.0.0alpha"),
        ]


class TestDependencyType(unittest.TestCase):
    """Ensure that the DependencyType dataclass converts to the correct strings"""

    def setUp(self) -> None:
        from addonmanager_metadata import DependencyType

        self.DependencyType = DependencyType

    def test_string_conversion_automatic(self):
        self.assertEqual(str(self.DependencyType.automatic), "automatic")

    def test_string_conversion_internal(self):
        self.assertEqual(str(self.DependencyType.internal), "internal")

    def test_string_conversion_addon(self):
        self.assertEqual(str(self.DependencyType.addon), "addon")

    def test_string_conversion_python(self):
        self.assertEqual(str(self.DependencyType.python), "python")


class TestUrlType(unittest.TestCase):
    """Ensure that the UrlType dataclass converts to the correct strings"""

    def setUp(self) -> None:
        from addonmanager_metadata import UrlType

        self.UrlType = UrlType

    def test_string_conversion_website(self):
        self.assertEqual(str(self.UrlType.website), "website")

    def test_string_conversion_repository(self):
        self.assertEqual(str(self.UrlType.repository), "repository")

    def test_string_conversion_bugtracker(self):
        self.assertEqual(str(self.UrlType.bugtracker), "bugtracker")

    def test_string_conversion_readme(self):
        self.assertEqual(str(self.UrlType.readme), "readme")

    def test_string_conversion_documentation(self):
        self.assertEqual(str(self.UrlType.documentation), "documentation")

    def test_string_conversion_discussion(self):
        self.assertEqual(str(self.UrlType.discussion), "discussion")


class TestMetadataAuxiliaryFunctions(unittest.TestCase):
    def test_get_first_supported_freecad_version_simple(self):
        from addonmanager_metadata import (
            Metadata,
            Version,
            get_first_supported_freecad_version,
        )

        expected_result = Version(from_string="0.20.2beta")
        metadata = self.given_metadata_with_freecadmin_set(expected_result)
        first_version = get_first_supported_freecad_version(metadata)
        self.assertEqual(expected_result, first_version)

    @staticmethod
    def given_metadata_with_freecadmin_set(min_version):
        from addonmanager_metadata import Metadata

        metadata = Metadata()
        metadata.freecadmin = min_version
        return metadata

    def test_get_first_supported_freecad_version_with_content(self):
        from addonmanager_metadata import (
            Metadata,
            Version,
            get_first_supported_freecad_version,
        )

        expected_result = Version(from_string="0.20.2beta")
        metadata = self.given_metadata_with_freecadmin_in_content(expected_result)
        first_version = get_first_supported_freecad_version(metadata)
        self.assertEqual(expected_result, first_version)

    @staticmethod
    def given_metadata_with_freecadmin_in_content(min_version):
        from addonmanager_metadata import Metadata, Version

        v_list = min_version.version_as_list
        metadata = Metadata()
        wb1 = Metadata()
        wb1.freecadmin = Version(from_list=[v_list[0] + 1, v_list[1], v_list[2], v_list[3]])
        wb2 = Metadata()
        wb2.freecadmin = Version(from_list=[v_list[0], v_list[1] + 1, v_list[2], v_list[3]])
        wb3 = Metadata()
        wb3.freecadmin = Version(from_list=[v_list[0], v_list[1], v_list[2] + 1, v_list[3]])
        m1 = Metadata()
        m1.freecadmin = min_version
        metadata.content = {"workbench": [wb1, wb2, wb3], "macro": [m1]}
        return metadata


class TestMetadataReader(unittest.TestCase):
    """Test reading metadata from XML"""

    def setUp(self) -> None:
        if "xml.etree.ElementTree" in sys.modules:
            sys.modules.pop("xml.etree.ElementTree")
        if "addonmanager_metadata" in sys.modules:
            sys.modules.pop("addonmanager_metadata")

    def tearDown(self) -> None:
        if "xml.etree.ElementTree" in sys.modules:
            sys.modules.pop("xml.etree.ElementTree")
        if "addonmanager_metadata" in sys.modules:
            sys.modules.pop("addonmanager_metadata")

    def test_from_file(self):
        from addonmanager_metadata import MetadataReader

        MetadataReader.from_bytes = Mock()
        with tempfile.NamedTemporaryFile(delete=False) as temp:
            temp.write(b"Some data")
            temp.close()
            MetadataReader.from_file(temp.name)
            self.assertTrue(MetadataReader.from_bytes.called)
            MetadataReader.from_bytes.assert_called_once_with(b"Some data")
            os.unlink(temp.name)

    @unittest.skip("Breaks other tests, needs to be fixed")
    def test_from_bytes(self):
        import xml.etree.ElementTree

        with unittest.mock.patch("xml.etree.ElementTree") as element_tree_mock:
            from addonmanager_metadata import MetadataReader

            MetadataReader._process_element_tree = Mock()
            MetadataReader.from_bytes(b"Some data")
            element_tree_mock.parse.assert_called_once_with(b"Some data")

    def test_process_element_tree(self):
        from addonmanager_metadata import MetadataReader

        MetadataReader._determine_namespace = Mock(return_value="")
        element_tree_mock = Mock()
        MetadataReader._create_node = Mock()
        MetadataReader._process_element_tree(element_tree_mock)
        MetadataReader._create_node.assert_called_once()

    def test_determine_namespace_found_full(self):
        from addonmanager_metadata import MetadataReader

        root = Mock()
        root.tag = "{https://wiki.freecad.org/Package_Metadata}package"
        found_ns = MetadataReader._determine_namespace(root)
        self.assertEqual(found_ns, "{https://wiki.freecad.org/Package_Metadata}")

    def test_determine_namespace_found_empty(self):
        from addonmanager_metadata import MetadataReader

        root = Mock()
        root.tag = "package"
        found_ns = MetadataReader._determine_namespace(root)
        self.assertEqual(found_ns, "")

    def test_determine_namespace_not_found(self):
        from addonmanager_metadata import MetadataReader

        root = Mock()
        root.find = Mock(return_value=False)
        with self.assertRaises(RuntimeError):
            MetadataReader._determine_namespace(root)

    def test_parse_child_element_simple_strings(self):
        from addonmanager_metadata import Metadata, MetadataReader

        tags = ["name", "date", "description", "icon", "classname", "subdirectory"]
        for tag in tags:
            with self.subTest(tag=tag):
                text = f"Test Data for {tag}"
                child = self.given_mock_tree_node(tag, text)
                mock_metadata = Metadata()
                MetadataReader._parse_child_element("", child, mock_metadata)
                self.assertEqual(mock_metadata.__dict__[tag], text)

    def test_parse_child_element_version(self):
        from addonmanager_metadata import Metadata, Version, MetadataReader

        mock_metadata = Metadata()
        child = self.given_mock_tree_node("version", "1.2.3")
        MetadataReader._parse_child_element("", child, mock_metadata)
        self.assertEqual(Version("1.2.3"), mock_metadata.version)

    def test_parse_child_element_version_bad(self):
        from addonmanager_metadata import Metadata, Version, MetadataReader

        mock_metadata = Metadata()
        child = self.given_mock_tree_node("version", "1-2-3")
        MetadataReader._parse_child_element("", child, mock_metadata)
        self.assertEqual(Version("0.0.0"), mock_metadata.version)

    def test_parse_child_element_lists_of_strings(self):
        from addonmanager_metadata import Metadata, MetadataReader

        tags = ["file", "tag"]
        for tag in tags:
            with self.subTest(tag=tag):
                mock_metadata = Metadata()
                expected_results = []
                for i in range(10):
                    text = f"Test {i} for {tag}"
                    expected_results.append(text)
                    child = self.given_mock_tree_node(tag, text)
                    MetadataReader._parse_child_element("", child, mock_metadata)
                self.assertEqual(len(mock_metadata.__dict__[tag]), 10)
                self.assertListEqual(mock_metadata.__dict__[tag], expected_results)

    def test_parse_child_element_lists_of_contacts(self):
        from addonmanager_metadata import Metadata, Contact, MetadataReader

        tags = ["maintainer", "author"]
        for tag in tags:
            with self.subTest(tag=tag):
                mock_metadata = Metadata()
                expected_results = []
                for i in range(10):
                    text = f"Test {i} for {tag}"
                    email = f"Email {i} for {tag}" if i % 2 == 0 else None
                    expected_results.append(Contact(name=text, email=email))
                    child = self.given_mock_tree_node(tag, text, {"email": email})
                    MetadataReader._parse_child_element("", child, mock_metadata)
                self.assertEqual(len(mock_metadata.__dict__[tag]), 10)
                self.assertListEqual(mock_metadata.__dict__[tag], expected_results)

    def test_parse_child_element_list_of_licenses(self):
        from addonmanager_metadata import Metadata, License, MetadataReader

        mock_metadata = Metadata()
        expected_results = []
        tag = "license"
        for i in range(10):
            text = f"Test {i} for {tag}"
            file = f"Filename {i} for {tag}" if i % 2 == 0 else None
            expected_results.append(License(name=text, file=file))
            child = self.given_mock_tree_node(tag, text, {"file": file})
            MetadataReader._parse_child_element("", child, mock_metadata)
        self.assertEqual(len(mock_metadata.__dict__[tag]), 10)
        self.assertListEqual(mock_metadata.__dict__[tag], expected_results)

    def test_parse_child_element_list_of_urls(self):
        from addonmanager_metadata import Metadata, Url, UrlType, MetadataReader

        mock_metadata = Metadata()
        expected_results = []
        tag = "url"
        for i in range(10):
            text = f"Test {i} for {tag}"
            url_type = UrlType(i % len(UrlType))
            type = str(url_type)
            branch = ""
            if type == "repository":
                branch = f"Branch {i} for {tag}"
            expected_results.append(Url(location=text, type=url_type, branch=branch))
            child = self.given_mock_tree_node(tag, text, {"type": type, "branch": branch})
            MetadataReader._parse_child_element("", child, mock_metadata)
        self.assertEqual(len(mock_metadata.__dict__[tag]), 10)
        self.assertListEqual(mock_metadata.__dict__[tag], expected_results)

    def test_parse_child_element_lists_of_dependencies(self):
        from addonmanager_metadata import (
            Metadata,
            Dependency,
            DependencyType,
            MetadataReader,
        )

        tags = ["depend", "conflict", "replace"]
        attributes = {
            "version_lt": "1.0.0",
            "version_lte": "1.0.0",
            "version_eq": "1.0.0",
            "version_gte": "1.0.0",
            "version_gt": "1.0.0",
            "condition": "$BuildVersionMajor<1",
            "optional": True,
        }

        for tag in tags:
            for attribute, attr_value in attributes.items():
                with self.subTest(tag=tag, attribute=attribute):
                    mock_metadata = Metadata()
                    expected_results = []
                    for i in range(10):
                        text = f"Test {i} for {tag}"
                        dependency_type = DependencyType(i % len(DependencyType))
                        dependency_type_str = str(dependency_type)
                        expected = Dependency(package=text, dependency_type=dependency_type)
                        expected.__dict__[attribute] = attr_value
                        expected_results.append(expected)
                        child = self.given_mock_tree_node(
                            tag,
                            text,
                            {"type": dependency_type_str, attribute: str(attr_value)},
                        )
                        MetadataReader._parse_child_element("", child, mock_metadata)
                    self.assertEqual(len(mock_metadata.__dict__[tag]), 10)
                    self.assertListEqual(mock_metadata.__dict__[tag], expected_results)

    def test_parse_child_element_ignore_unknown_tag(self):
        from addonmanager_metadata import Metadata, MetadataReader

        tag = "invalid_tag"
        text = "Shouldn't matter"
        child = self.given_mock_tree_node(tag, text)
        mock_metadata = Metadata()
        MetadataReader._parse_child_element("", child, mock_metadata)
        self.assertNotIn(tag, mock_metadata.__dict__)

    def test_parse_child_element_versions(self):
        from addonmanager_metadata import Metadata, Version, MetadataReader

        tags = ["version", "freecadmin", "freecadmax", "pythonmin"]
        for tag in tags:
            with self.subTest(tag=tag):
                mock_metadata = Metadata()
                text = "3.4.5beta"
                child = self.given_mock_tree_node(tag, text)
                MetadataReader._parse_child_element("", child, mock_metadata)
                self.assertEqual(mock_metadata.__dict__[tag], Version(from_string=text))

    def given_mock_tree_node(self, tag, text, attributes=None):
        class MockTreeNode:
            def __init__(self):
                self.tag = tag
                self.text = text
                self.attrib = attributes if attributes is not None else []

        return MockTreeNode()

    def test_parse_content_valid(self):
        from addonmanager_metadata import MetadataReader

        valid_content_items = ["workbench", "macro", "preferencepack"]
        MetadataReader._create_node = Mock()
        for content_type in valid_content_items:
            with self.subTest(content_type=content_type):
                tree_mock = [self.given_mock_tree_node(content_type, None)]
                metadata_mock = Mock()
                MetadataReader._parse_content("", metadata_mock, tree_mock)
                MetadataReader._create_node.assert_called_once()
                MetadataReader._create_node.reset_mock()

    def test_parse_content_invalid(self):
        from addonmanager_metadata import MetadataReader

        MetadataReader._create_node = Mock()
        content_item = "no_such_content_type"
        tree_mock = [self.given_mock_tree_node(content_item, None)]
        metadata_mock = Mock()
        MetadataReader._parse_content("", metadata_mock, tree_mock)
        MetadataReader._create_node.assert_not_called()


class TestMetadataReaderIntegration(unittest.TestCase):
    """Full-up tests of the MetadataReader class (no mocking)."""

    def setUp(self) -> None:
        self.test_data_dir = os.path.join(os.path.dirname(__file__), "..", "data")
        remove_list = []
        for key in sys.modules:
            if "addonmanager_metadata" in key:
                remove_list.append(key)
        for key in remove_list:
            print(f"Removing {key}")
            sys.modules.pop(key)

    def test_loading_simple_metadata_file(self):
        from addonmanager_metadata import (
            Contact,
            Dependency,
            License,
            MetadataReader,
            Url,
            UrlType,
            Version,
        )

        filename = os.path.join(self.test_data_dir, "good_package.xml")
        metadata = MetadataReader.from_file(filename)
        self.assertEqual("Test Workbench", metadata.name)
        self.assertEqual("A package.xml file for unit testing.", metadata.description)
        self.assertEqual(Version("1.0.1"), metadata.version)
        self.assertEqual("2022-01-07", metadata.date)
        self.assertEqual("Resources/icons/PackageIcon.svg", metadata.icon)
        self.assertListEqual([License(name="LGPLv2.1", file="LICENSE")], metadata.license)
        self.assertListEqual(
            [Contact(name="FreeCAD Developer", email="developer@freecad.org")],
            metadata.maintainer,
        )
        self.assertListEqual(
            [
                Url(
                    location="https://github.com/chennes/FreeCAD-Package",
                    type=UrlType.repository,
                    branch="main",
                ),
                Url(
                    location="https://github.com/chennes/FreeCAD-Package/blob/main/README.md",
                    type=UrlType.readme,
                ),
            ],
            metadata.url,
        )
        self.assertListEqual(["Tag0", "Tag1"], metadata.tag)
        self.assertIn("workbench", metadata.content)
        self.assertEqual(len(metadata.content["workbench"]), 1)
        wb_metadata = metadata.content["workbench"][0]
        self.assertEqual("MyWorkbench", wb_metadata.classname)
        self.assertEqual("./", wb_metadata.subdirectory)
        self.assertListEqual(["TagA", "TagB", "TagC"], wb_metadata.tag)

    def test_multiple_workbenches(self):
        from addonmanager_metadata import MetadataReader

        filename = os.path.join(self.test_data_dir, "workbench_only.xml")
        metadata = MetadataReader.from_file(filename)
        self.assertIn("workbench", metadata.content)
        self.assertEqual(len(metadata.content["workbench"]), 3)
        expected_wb_classnames = [
            "MyFirstWorkbench",
            "MySecondWorkbench",
            "MyThirdWorkbench",
        ]
        for wb in metadata.content["workbench"]:
            self.assertIn(wb.classname, expected_wb_classnames)
            expected_wb_classnames.remove(wb.classname)
        self.assertEqual(len(expected_wb_classnames), 0)

    def test_multiple_macros(self):
        from addonmanager_metadata import MetadataReader

        filename = os.path.join(self.test_data_dir, "macro_only.xml")
        metadata = MetadataReader.from_file(filename)
        self.assertIn("macro", metadata.content)
        self.assertEqual(len(metadata.content["macro"]), 2)
        expected_wb_files = ["MyMacro.FCStd", "MyOtherMacro.FCStd"]
        for wb in metadata.content["macro"]:
            self.assertIn(wb.file[0], expected_wb_files)
            expected_wb_files.remove(wb.file[0])
        self.assertEqual(len(expected_wb_files), 0)

    def test_multiple_preference_packs(self):
        from addonmanager_metadata import MetadataReader

        filename = os.path.join(self.test_data_dir, "prefpack_only.xml")
        metadata = MetadataReader.from_file(filename)
        self.assertIn("preferencepack", metadata.content)
        self.assertEqual(len(metadata.content["preferencepack"]), 3)
        expected_packs = ["MyFirstPack", "MySecondPack", "MyThirdPack"]
        for wb in metadata.content["preferencepack"]:
            self.assertIn(wb.name, expected_packs)
            expected_packs.remove(wb.name)
        self.assertEqual(len(expected_packs), 0)

    def test_content_combination(self):
        from addonmanager_metadata import MetadataReader

        filename = os.path.join(self.test_data_dir, "combination.xml")
        metadata = MetadataReader.from_file(filename)
        self.assertIn("preferencepack", metadata.content)
        self.assertEqual(len(metadata.content["preferencepack"]), 1)
        self.assertIn("macro", metadata.content)
        self.assertEqual(len(metadata.content["macro"]), 1)
        self.assertIn("workbench", metadata.content)
        self.assertEqual(len(metadata.content["workbench"]), 1)


if __name__ == "__main__":
    unittest.main()



#***************************************************************************
#*   Copyright (c) 2014 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

import pathlib
import FreeCAD, Part
from PySide import QtCore
from fcgear import involute
from fcgear import fcgear

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtGui

__title__="PartDesign InvoluteGearObject management"
__author__ = "Juergen Riegel"
__url__ = "http://www.freecad.org"


def makeInvoluteGear(name):
    '''makeInvoluteGear(name): makes an InvoluteGear'''
    obj = FreeCAD.ActiveDocument.addObject("Part::Part2DObjectPython",name)
    _InvoluteGear(obj)
    if FreeCAD.GuiUp:
        _ViewProviderInvoluteGear(obj.ViewObject)
    #FreeCAD.ActiveDocument.recompute()
    if FreeCAD.GuiUp:
        body=FreeCADGui.ActiveDocument.ActiveView.getActiveObject("pdbody")
        part=FreeCADGui.ActiveDocument.ActiveView.getActiveObject("part")
        if body:
            body.Group=body.Group+[obj]
        elif part:
            part.Group=part.Group+[obj]
    return obj


class _CommandInvoluteGear:
    "GUI command to create an InvoluteGear"
    def GetResources(self):
        return {'Pixmap'  : 'PartDesign_InternalExternalGear',
                'MenuText': QtCore.QT_TRANSLATE_NOOP("PartDesign_InvoluteGear","Involute gear..."),
                'Accel': "",
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("PartDesign_InvoluteGear","Creates or edit the involute gear definition.")}

    def Activated(self):
        FreeCAD.ActiveDocument.openTransaction("Create involute gear")
        FreeCADGui.addModule("InvoluteGearFeature")
        FreeCADGui.doCommand("InvoluteGearFeature.makeInvoluteGear('InvoluteGear')")
        FreeCADGui.doCommand("Gui.activeDocument().setEdit(App.ActiveDocument.ActiveObject.Name,0)")

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


class _InvoluteGear:
    "The InvoluteGear object"
    def __init__(self,obj):
        self.Type = "InvoluteGear"
        self._ensure_properties(obj, is_restore=False)
        obj.Proxy = self

    def onDocumentRestored(self, obj):
        """hook used to migrate older versions of this object"""
        self._ensure_properties(obj, is_restore=True)

    def _ensure_properties(self, obj, is_restore):
        def ensure_property(type_, name, doc, default):
            if not hasattr(obj, name):
                obj.addProperty(type_, name, "Gear", doc)
                if callable(default):
                    setattr(obj, name, default())
                else:
                    setattr(obj, name, default)

        # for details about the property's docstring translation,
        # see https://tracker.freecad.org/view.php?id=2524
        ensure_property("App::PropertyInteger", "NumberOfTeeth",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property", "Number of gear teeth"),
            default=26)
        ensure_property("App::PropertyLength", "Modules",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property", "Modules of the gear"),
            default="2.5 mm")
        ensure_property("App::PropertyAngle", "PressureAngle",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property", "Pressure angle of gear teeth"),
            default="20 deg")
        ensure_property("App::PropertyBool", "HighPrecision",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property",
                "True=2 curves with each 3 control points False=1 curve with 4 control points."),
            default=True)
        ensure_property("App::PropertyBool", "ExternalGear",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property", "True=external Gear False=internal Gear"),
            default=True)
        ensure_property("App::PropertyFloat", "AddendumCoefficient",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property",
                "The height of the tooth from the pitch circle up to its tip, normalized by the module."),
            default=lambda: 1.0 if obj.ExternalGear else 0.6)
        ensure_property("App::PropertyFloat","DedendumCoefficient",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property",
                "The height of the tooth from the pitch circle down to its root, normalized by the module."),
            default=1.25)
        ensure_property("App::PropertyFloat","RootFilletCoefficient",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property",
                "The radius of the fillet at the root of the tooth, normalized by the module."),
            default=lambda: 0.375 if is_restore else 0.38)
        ensure_property("App::PropertyFloat","ProfileShiftCoefficient",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property",
                "The distance by which the reference profile is shifted outwards, normalized by the module."),
            default=0.0)

    def execute(self,obj):
        w = fcgear.FCWireBuilder()
        generator_func = involute.CreateExternalGear if obj.ExternalGear else involute.CreateInternalGear
        generator_func(w, obj.Modules.Value, obj.NumberOfTeeth, obj.PressureAngle.Value,
            split=obj.HighPrecision, addCoeff=obj.AddendumCoefficient, dedCoeff=obj.DedendumCoefficient,
            filletCoeff=obj.RootFilletCoefficient, shiftCoeff=obj.ProfileShiftCoefficient)
        gearw = Part.Wire([o.toShape() for o in w.wire])
        obj.Shape = gearw
        obj.positionBySupport()
        return


class _ViewProviderInvoluteGear:
    "A View Provider for the InvoluteGear object"

    def __init__(self,vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/PartDesign_InternalExternalGear.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def setEdit(self,vobj,mode):
        taskd = _InvoluteGearTaskPanel(self.Object,mode)
        taskd.obj = vobj.Object
        taskd.update()
        FreeCADGui.Control.showDialog(taskd)
        return True

    def unsetEdit(self,vobj,mode):
        FreeCADGui.Control.closeDialog()
        return

    def dumps(self):
        return None

    def loads(self,state):
        return None


class _InvoluteGearTaskPanel:
    '''The editmode TaskPanel for InvoluteGear objects'''
    def __init__(self,obj,mode):
        self.obj = obj

        self.form=FreeCADGui.PySideUic.loadUi(str(pathlib.Path(__file__).with_suffix(".ui")))
        self.form.setWindowIcon(QtGui.QIcon(":/icons/PartDesign_InternalExternalGear.svg"))
        self.assignToolTipsFromPropertyDocs()

        def assignValue(property_name, fitView=False):
            """Returns a function that takes a single value and assigns it to the given property"""
            def assigner(value):
                setattr(self.obj, property_name, value)
                self.obj.Proxy.execute(self.obj)
                if fitView:
                    FreeCAD.Gui.SendMsgToActiveView("ViewFit")
            return assigner

        def assignIndexAsBool(property_name):
            """Variant of assignValue that transforms the index of a Yes/No Combobox to a bool."""
            assigner = assignValue(property_name)
            def transformingAssigner(value):
                assigner(True if value == 0 else False)
            return transformingAssigner

        self.form.Quantity_Modules.valueChanged.connect(assignValue("Modules", fitView=True))
        self.form.Quantity_PressureAngle.valueChanged.connect(assignValue("PressureAngle"))
        self.form.spinBox_NumberOfTeeth.valueChanged.connect(assignValue("NumberOfTeeth", fitView=True))
        self.form.comboBox_HighPrecision.currentIndexChanged.connect(assignIndexAsBool("HighPrecision"))
        self.form.comboBox_ExternalGear.currentIndexChanged.connect(assignIndexAsBool("ExternalGear"))
        self.form.doubleSpinBox_Addendum.valueChanged.connect(assignValue("AddendumCoefficient"))
        self.form.doubleSpinBox_Dedendum.valueChanged.connect(assignValue("DedendumCoefficient"))
        self.form.doubleSpinBox_RootFillet.valueChanged.connect(assignValue("RootFilletCoefficient"))
        self.form.doubleSpinBox_ProfileShift.valueChanged.connect(assignValue("ProfileShiftCoefficient"))

        self.update()

        if mode == 0: # fresh created
            self.obj.Proxy.execute(self.obj)  # calculate once
            FreeCAD.Gui.SendMsgToActiveView("ViewFit")

    def assignToolTipsFromPropertyDocs(self):
        def assign(property_name, *widgets):
            doc = self.obj.getDocumentationOfProperty(property_name)
            translated_doc = QtGui.QApplication.translate("App::Property", doc)
            for w in widgets:
                w.setToolTip(translated_doc)

        # we assign the tool tip to both, the label and the input field, for user convenience
        assign("Modules", self.form.Quantity_Modules, self.form.label_Modules)
        assign("PressureAngle", self.form.Quantity_PressureAngle, self.form.label_PressureAngle)
        assign("NumberOfTeeth", self.form.spinBox_NumberOfTeeth, self.form.label_NumberOfTeeth)
        assign("HighPrecision", self.form.comboBox_HighPrecision, self.form.label_HighPrecision)
        assign("ExternalGear", self.form.comboBox_ExternalGear, self.form.label_ExternalGear)
        assign("AddendumCoefficient", self.form.doubleSpinBox_Addendum, self.form.label_Addendum)
        assign("DedendumCoefficient", self.form.doubleSpinBox_Dedendum, self.form.label_Dedendum)
        assign("RootFilletCoefficient", self.form.doubleSpinBox_RootFillet, self.form.label_RootFillet)
        assign("ProfileShiftCoefficient", self.form.doubleSpinBox_ProfileShift, self.form.label_ProfileShift)

    def changeEvent(self, event):
        if event == QtCore.QEvent.LanguageChange:
            self.assignToolTipsFromPropertyDocs()

    def transferTo(self):
        "Transfer from the dialog to the object"
        self.obj.NumberOfTeeth  = self.form.spinBox_NumberOfTeeth.value()
        self.obj.Modules        = self.form.Quantity_Modules.text()
        self.obj.PressureAngle  = self.form.Quantity_PressureAngle.text()
        self.obj.HighPrecision = True if self.form.comboBox_HighPrecision.currentIndex() == 0 else False
        self.obj.ExternalGear = True if self.form.comboBox_ExternalGear.currentIndex() == 0 else False
        self.obj.AddendumCoefficient = self.form.doubleSpinBox_Addendum.value()
        self.obj.DedendumCoefficient = self.form.doubleSpinBox_Dedendum.value()
        self.obj.RootFilletCoefficient = self.form.doubleSpinBox_RootFillet.value()
        self.obj.ProfileShiftCoefficient = self.form.doubleSpinBox_ProfileShift.value()

    def transferFrom(self):
        "Transfer from the object to the dialog"
        self.form.spinBox_NumberOfTeeth.setValue(self.obj.NumberOfTeeth)
        self.form.Quantity_Modules.setText(self.obj.Modules.UserString)
        self.form.Quantity_PressureAngle.setText(self.obj.PressureAngle.UserString)
        self.form.comboBox_HighPrecision.setCurrentIndex(0 if self.obj.HighPrecision else 1)
        self.form.comboBox_ExternalGear.setCurrentIndex(0 if self.obj.ExternalGear else 1)
        self.form.doubleSpinBox_Addendum.setValue(self.obj.AddendumCoefficient)
        self.form.doubleSpinBox_Dedendum.setValue(self.obj.DedendumCoefficient)
        self.form.doubleSpinBox_RootFillet.setValue(self.obj.RootFilletCoefficient)
        self.form.doubleSpinBox_ProfileShift.setValue(self.obj.ProfileShiftCoefficient)

    def getStandardButtons(self):
        return int(QtGui.QDialogButtonBox.Ok) | int(QtGui.QDialogButtonBox.Cancel)| int(QtGui.QDialogButtonBox.Apply)

    def clicked(self,button):
        if button == QtGui.QDialogButtonBox.Apply:
            self.transferTo()
            self.obj.Proxy.execute(self.obj)

    def update(self):
        'fills the widgets'
        self.transferFrom()

    def accept(self):
        self.transferTo()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()

    def reject(self):
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCAD.ActiveDocument.abortTransaction()


if FreeCAD.GuiUp:
    FreeCADGui.addCommand('PartDesign_InvoluteGear',_CommandInvoluteGear())



#! python
# -*- coding: utf-8 -*-
# (c) 2011 Adrian Przekwas LGPL

from __future__ import division # allows floating point division from integers
import FreeCAD, Part
from FreeCAD import Base

class MySpring:
   def __init__(self, obj):
      ''' Add the properties: Pitch, Diameter, Height, BarDiameter '''
      obj.addProperty("App::PropertyLength", "Pitch", "MySpring", "Pitch of the helix").Pitch = 5.0
      obj.addProperty("App::PropertyLength", "Diameter", "MySpring", "Diameter of the helix").Diameter = 6.0
      obj.addProperty("App::PropertyLength", "Height", "MySpring", "Height of the helix").Height = 30.0
      obj.addProperty("App::PropertyLength", "BarDiameter", "MySpring", "Diameter of the bar").BarDiameter = 3.0
      obj.Proxy = self

   def onChanged(self, fp, prop):
      if prop == "Pitch" or prop == "Diameter" or prop == "Height" or prop == "BarDiameter":
         self.execute(fp)

   def execute(self, fp):
      pitch = fp.Pitch
      radius = fp.Diameter/2
      height = fp.Height
      barradius = fp.BarDiameter/2
      myhelix = Part.makeHelix(pitch, height, radius)
      g = myhelix.Edges[0].Curve
      c = Part.Circle()
      c.Center = g.value(0)  # start point of the helix
      c.Axis = (0, 1, 0)
      c.Radius = barradius
      p = c.toShape()
      section = Part.Wire([p])
      makeSolid = True
      isFrenet = True
      myspring = Part.Wire(myhelix).makePipeShell([section], makeSolid, isFrenet)
      fp.Shape = myspring

def makeMySpring():
   doc = FreeCAD.activeDocument()
   if doc is None:
      doc = FreeCAD.newDocument()
   spring = doc.addObject("Part::FeaturePython", "My_Spring")
   spring.Label = "My Spring"
   MySpring(spring)
   spring.ViewObject.Proxy = 0
   doc.recompute()


if __name__ == "__main__":
   makeMySpring()



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver object Elmer"
__author__ = "Markus Hovorka"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import glob
import os

import FreeCAD

from . import tasks
from .equations import deformation
from .equations import elasticity
from .equations import electricforce
from .equations import electrostatic
from .equations import flow
from .equations import flux
from .equations import heat
from .equations import magnetodynamic
from .equations import magnetodynamic2D
from .. import run
from .. import solverbase
from femtools import femutils

if FreeCAD.GuiUp:
    import FemGui

COORDINATE_SYSTEM = ["Cartesian", "Cartesian 1D", "Cartesian 2D", "Cartesian 3D",
                     "Polar 2D", "Polar 3D",
                     "Cylindric", "Cylindric Symmetric",
                     "Axi Symmetric"]
SIMULATION_TYPE = ["Scanning", "Steady State", "Transient"]


def create(doc, name="ElmerSolver"):
    return femutils.createObject(
        doc, name, Proxy, ViewProxy)


class Proxy(solverbase.Proxy):
    """Proxy for FemSolverElmers Document Object."""

    Type = "Fem::SolverElmer"

    _EQUATIONS = {
        "Deformation": deformation,
        "Elasticity": elasticity,
        "Electrostatic": electrostatic,
        "Flux": flux,
        "Electricforce": electricforce,
        "Flow": flow,
        "Heat": heat,
        "Magnetodynamic": magnetodynamic,
        "Magnetodynamic2D": magnetodynamic2D,
    }

    def __init__(self, obj):
        super(Proxy, self).__init__(obj)

        obj.addProperty(
            "App::PropertyEnumeration",
            "CoordinateSystem",
            "Coordinate System",
            ""
        )
        obj.CoordinateSystem = COORDINATE_SYSTEM
        obj.CoordinateSystem = "Cartesian"

        obj.addProperty(
            "App::PropertyIntegerConstraint",
            "BDFOrder",
            "Timestepping",
            "Order of time stepping method 'BDF'"
        )
        # according to the Elmer manual recommended is order 2
        # possible ranage is 1 - 5
        obj.BDFOrder = (2, 1, 5, 1)

        obj.addProperty(
            "App::PropertyIntegerList",
            "OutputIntervals",
            "Timestepping",
            "After how many time steps a result file is output"
        )
        obj.OutputIntervals = [1]

        obj.addProperty(
            "App::PropertyIntegerList",
            "TimestepIntervals",
            "Timestepping",
            (
                "List of times if 'Simulation Type'\n"
                "is either 'Scanning' or 'Transient'"
            )
        )
        obj.addProperty(
            "App::PropertyFloatList",
            "TimestepSizes",
            "Timestepping",
            (
                "List of time steps sizes if 'Simulation Type'\n"
                "is either 'Scanning' or 'Transient'"
            )
        )
        # there is no universal default, it all depends on the analysis, however
        # we have to set something and set 10 seconds in steps of 0.1s
        obj.TimestepIntervals = [100]
        obj.TimestepSizes = [0.1]

        obj.addProperty(
            "App::PropertyEnumeration",
            "SimulationType",
            "Type",
            ""
        )
        obj.SimulationType = SIMULATION_TYPE
        obj.SimulationType = "Steady State"

        obj.addProperty(
            "App::PropertyInteger",
            "SteadyStateMaxIterations",
            "Type",
            "Maximal steady state iterations"
        )
        obj.SteadyStateMaxIterations = 1

        obj.addProperty(
            "App::PropertyInteger",
            "SteadyStateMinIterations",
            "Type",
            "Minimal steady state iterations"
        )
        obj.SteadyStateMinIterations = 0

        obj.addProperty(
            "App::PropertyLink",
            "ElmerResult",
            "Base",
            "",
            4 | 8
        )

        obj.addProperty(
            "App::PropertyLinkList",
            "ElmerTimeResults",
            "Base",
            "",
            4 | 8
        )

        obj.addProperty(
            "App::PropertyLink",
            "ElmerOutput",
            "Base",
            "",
            4 | 8
        )

    def createMachine(self, obj, directory, testmode=False):
        return run.Machine(
            solver=obj, directory=directory,
            check=tasks.Check(),
            prepare=tasks.Prepare(),
            solve=tasks.Solve(),
            results=tasks.Results(),
            testmode=testmode)

    def createEquation(self, doc, eqId):
        return self._EQUATIONS[eqId].create(doc)

    def isSupported(self, eqId):
        return eqId in self._EQUATIONS

    def editSupported(self):
        return True

    def edit(self, directory):
        pattern = os.path.join(directory, "case.sif")
        FreeCAD.Console.PrintMessage("{}\n".format(pattern))
        f = glob.glob(pattern)[0]
        FemGui.open(f)


class ViewProxy(solverbase.ViewProxy):
    """Proxy for FemSolverElmers View Provider."""

    def getIcon(self):
        return ":/icons/FEM_SolverElmer.svg"

##  @}



# ***************************************************************************
# *   Copyright (c) 2017 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver object CalculiX"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package SolverCalculix
#  \ingroup FEM

import glob
import os

import FreeCAD

from . import tasks
from .. import run
from .. import solverbase
from femtools import femutils

if FreeCAD.GuiUp:
    import FemGui

ANALYSIS_TYPES = ["static", "frequency", "thermomech", "check", "buckling"]


def create(doc, name="SolverCalculiX"):
    return femutils.createObject(
        doc, name, Proxy, ViewProxy)


class Proxy(solverbase.Proxy):
    """The Fem::FemSolver's Proxy python type, add solver specific properties
    """

    Type = "Fem::SolverCalculix"

    def __init__(self, obj):
        super(Proxy, self).__init__(obj)
        obj.Proxy = self
        ccx_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/Ccx")
        add_attributes(obj, ccx_prefs)

    def onDocumentRestored(self, obj):
        ccx_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/Ccx")
        # since it is needed for the ccxtools solver too
        # the method is implemented outside of the class
        # thus we need to pass the prefs
        on_restore_of_document(obj, ccx_prefs)

    def createMachine(self, obj, directory, testmode=False):
        return run.Machine(
            solver=obj, directory=directory,
            check=tasks.Check(),
            prepare=tasks.Prepare(),
            solve=tasks.Solve(),
            results=tasks.Results(),
            testmode=testmode)

    def editSupported(self):
        return True

    def edit(self, directory):
        pattern = os.path.join(directory, "*.inp")
        FreeCAD.Console.PrintMessage("{}\n".format(pattern))
        f = glob.glob(pattern)[0]
        FemGui.open(f)

    def execute(self, obj):
        return


class ViewProxy(solverbase.ViewProxy):
    pass


# ************************************************************************************************
# helper
# these methods are outside of the class to be able
# to use them from framework solver and ccxtools solver


def on_restore_of_document(obj, ccx_prefs):

    # ANALYSIS_TYPES
    # They have been extended. If file was saved with a old FC version
    # not all enum types are available, because they are saved in the FC file
    # thus refresh the list of known ANALYSIS_TYPES
    # print("onRestoredFromSuperClass")
    # print(obj.AnalysisType)
    # print(obj.getEnumerationsOfProperty("AnalysisType"))

    temp_analysis_type = obj.AnalysisType
    # self.add_properties(obj)
    obj.AnalysisType = ANALYSIS_TYPES
    if temp_analysis_type in ANALYSIS_TYPES:
        obj.AnalysisType = temp_analysis_type
    else:
        FreeCAD.Console.PrintWarning(
            "Analysis type {} not found. Standard is used.\n"
            .format(temp_analysis_type)
        )
        analysis_type = ccx_prefs.GetInt("AnalysisType", 0)
        obj.AnalysisType = ANALYSIS_TYPES[analysis_type]

    # add missing properties
    # for example BucklingFactors will be added
    # for all files created before buckle analysis was introduced
    add_attributes(obj, ccx_prefs)


def add_attributes(obj, ccx_prefs):

    if not hasattr(obj, "AnalysisType"):
        obj.addProperty(
            "App::PropertyEnumeration",
            "AnalysisType",
            "Fem",
            "Type of the analysis"
        )
        obj.AnalysisType = ANALYSIS_TYPES
        analysis_type = ccx_prefs.GetInt("AnalysisType", 0)
        obj.AnalysisType = ANALYSIS_TYPES[analysis_type]

    if not hasattr(obj, "GeometricalNonlinearity"):
        choices_geom_nonlinear = ["linear", "nonlinear"]
        obj.addProperty(
            "App::PropertyEnumeration",
            "GeometricalNonlinearity",
            "Fem",
            "Set geometrical nonlinearity"
        )
        obj.GeometricalNonlinearity = choices_geom_nonlinear
        nonlinear_geom = ccx_prefs.GetBool("NonlinearGeometry", False)
        if nonlinear_geom is True:
            obj.GeometricalNonlinearity = choices_geom_nonlinear[1]  # nonlinear
        else:
            obj.GeometricalNonlinearity = choices_geom_nonlinear[0]  # linear

    if not hasattr(obj, "MaterialNonlinearity"):
        choices_material_nonlinear = ["linear", "nonlinear"]
        obj.addProperty(
            "App::PropertyEnumeration",
            "MaterialNonlinearity",
            "Fem",
            "Set material nonlinearity (needs geometrical nonlinearity)"
        )
        obj.MaterialNonlinearity = choices_material_nonlinear
        obj.MaterialNonlinearity = choices_material_nonlinear[0]

    if not hasattr(obj, "EigenmodesCount"):
        obj.addProperty(
            "App::PropertyIntegerConstraint",
            "EigenmodesCount",
            "Fem",
            "Number of modes for frequency calculations"
        )
        noem = ccx_prefs.GetInt("EigenmodesCount", 10)
        obj.EigenmodesCount = (noem, 1, 100, 1)

    if not hasattr(obj, "EigenmodeLowLimit"):
        obj.addProperty(
            "App::PropertyFloatConstraint",
            "EigenmodeLowLimit",
            "Fem",
            "Low frequency limit for eigenmode calculations"
        )
        ell = ccx_prefs.GetFloat("EigenmodeLowLimit", 0.0)
        obj.EigenmodeLowLimit = (ell, 0.0, 1000000.0, 10000.0)

    if not hasattr(obj, "EigenmodeHighLimit"):
        obj.addProperty(
            "App::PropertyFloatConstraint",
            "EigenmodeHighLimit",
            "Fem",
            "High frequency limit for eigenmode calculations"
        )
        ehl = ccx_prefs.GetFloat("EigenmodeHighLimit", 1000000.0)
        obj.EigenmodeHighLimit = (ehl, 0.0, 1000000.0, 10000.0)

    if not hasattr(obj, "IterationsThermoMechMaximum"):
        help_string_IterationsThermoMechMaximum = (
            "Maximum Number of thermo mechanical iterations "
            "in each time step before stopping jobs"
        )
        obj.addProperty(
            "App::PropertyIntegerConstraint",
            "IterationsThermoMechMaximum",
            "Fem",
            help_string_IterationsThermoMechMaximum
        )
        niter = ccx_prefs.GetInt("AnalysisMaxIterations", 200)
        obj.IterationsThermoMechMaximum = niter

    if not hasattr(obj, "BucklingFactors"):
        obj.addProperty(
            "App::PropertyIntegerConstraint",
            "BucklingFactors",
            "Fem",
            "Calculates the lowest buckling modes to the corresponding buckling factors"
        )
        bckl = ccx_prefs.GetInt("BucklingFactors", 1)
        obj.BucklingFactors = bckl

    if not hasattr(obj, "TimeInitialStep"):
        obj.addProperty(
            "App::PropertyFloatConstraint",
            "TimeInitialStep",
            "Fem",
            "Initial time steps"
        )
        ini = ccx_prefs.GetFloat("AnalysisTimeInitialStep", 1.0)
        obj.TimeInitialStep = ini

    if not hasattr(obj, "TimeEnd"):
        obj.addProperty(
            "App::PropertyFloatConstraint",
            "TimeEnd",
            "Fem",
            "End time analysis"
        )
        eni = ccx_prefs.GetFloat("AnalysisTime", 1.0)
        obj.TimeEnd = eni

    if not hasattr(obj, "ThermoMechSteadyState"):
        obj.addProperty(
            "App::PropertyBool",
            "ThermoMechSteadyState",
            "Fem",
            "Choose between steady state thermo mech or transient thermo mech analysis"
        )
        sted = ccx_prefs.GetBool("StaticAnalysis", True)
        obj.ThermoMechSteadyState = sted

    if not hasattr(obj, "IterationsControlParameterTimeUse"):
        obj.addProperty(
            "App::PropertyBool",
            "IterationsControlParameterTimeUse",
            "Fem",
            "Use the user defined time incrementation control parameter"
        )
        use_non_ccx_iterations_param = ccx_prefs.GetInt("UseNonCcxIterationParam", False)
        obj.IterationsControlParameterTimeUse = use_non_ccx_iterations_param

    if not hasattr(obj, "SplitInputWriter"):
        obj.addProperty(
            "App::PropertyBool",
            "SplitInputWriter",
            "Fem",
            "Split writing of ccx input file"
        )
        split = ccx_prefs.GetBool("SplitInputWriter", False)
        obj.SplitInputWriter = split

    if not hasattr(obj, "IterationsControlParameterIter"):
        control_parameter_iterations = (
            "{I_0},{I_R},{I_P},{I_C},{I_L},{I_G},{I_S},{I_A},{I_J},{I_T}".format(
                I_0=4,
                I_R=8,
                I_P=9,
                I_C=200,  # ccx default = 16
                I_L=10,
                I_G=400,  # ccx default = 4
                I_S="",
                I_A=200,  # ccx default = 5
                I_J="",
                I_T="",
            )
        )
        obj.addProperty(
            "App::PropertyString",
            "IterationsControlParameterIter",
            "Fem",
            "User defined time incrementation iterations control parameter"
        )
        obj.IterationsControlParameterIter = control_parameter_iterations

    if not hasattr(obj, "IterationsControlParameterCutb"):
        control_parameter_cutback = (
            "{D_f},{D_C},{D_B},{D_A},{D_S},{D_H},{D_D},{W_G}".format(
                D_f=0.25,
                D_C=0.5,
                D_B=0.75,
                D_A=0.85,
                D_S="",
                D_H="",
                D_D=1.5,
                W_G="",
            )
        )
        obj.addProperty(
            "App::PropertyString",
            "IterationsControlParameterCutb",
            "Fem",
            "User defined time incrementation cutbacks control parameter"
        )
        obj.IterationsControlParameterCutb = control_parameter_cutback

    if not hasattr(obj, "IterationsUserDefinedIncrementations"):
        stringIterationsUserDefinedIncrementations = (
            "Set to True to switch off the ccx automatic incrementation completely "
            "(ccx parameter DIRECT). Use with care. Analysis may not converge!"
        )
        obj.addProperty(
            "App::PropertyBool",
            "IterationsUserDefinedIncrementations",
            "Fem",
            stringIterationsUserDefinedIncrementations
        )
        obj.IterationsUserDefinedIncrementations = False

    if not hasattr(obj, "IterationsUserDefinedTimeStepLength"):
        help_string_IterationsUserDefinedTimeStepLength = (
            "Set to True to use the user defined time steps. "
            "The time steps are set with TimeInitialStep and TimeEnd"
        )
        obj.addProperty(
            "App::PropertyBool",
            "IterationsUserDefinedTimeStepLength",
            "Fem",
            help_string_IterationsUserDefinedTimeStepLength
        )
        obj.IterationsUserDefinedTimeStepLength = False

    if not hasattr(obj, "MatrixSolverType"):
        known_ccx_solver_types = [
            "default",
            "spooles",
            "iterativescaling",
            "iterativecholesky"
        ]
        obj.addProperty(
            "App::PropertyEnumeration",
            "MatrixSolverType",
            "Fem",
            "Type of solver to use"
        )
        obj.MatrixSolverType = known_ccx_solver_types
        solver_type = ccx_prefs.GetInt("Solver", 0)
        obj.MatrixSolverType = known_ccx_solver_types[solver_type]

    if not hasattr(obj, "BeamShellResultOutput3D"):
        obj.addProperty(
            "App::PropertyBool",
            "BeamShellResultOutput3D",
            "Fem",
            "Output 3D results for 1D and 2D analysis "
        )
        dimout = ccx_prefs.GetBool("BeamShellOutput", False)
        obj.BeamShellResultOutput3D = dimout


"""
Should there be some equation object for Calculix too?

Necessarily yes! The properties GeometricalNonlinearity,
MaterialNonlinearity, ThermoMechSteadyState might be moved
to the appropriate equation.

Furthermore the material Category should not be used in writer.
See common material object for more information. The equation
should used instead to get this information needed in writer.
"""



# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create Layer objects."""
## @package gui_layers
# \ingroup draftguitools
# \brief Provides GUI tools to create Layer objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import os
import FreeCAD
import FreeCADGui as Gui
import Draft
import Draft_rc
import draftguitools.gui_base as gui_base

from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
bool(Draft_rc.__name__)


def getColorIcon(color):

    "returns a QtGui.QIcon from a color 3-float tuple"

    from PySide import QtCore,QtGui
    c = QtGui.QColor(int(color[0]*255),int(color[1]*255),int(color[2]*255))
    im = QtGui.QImage(48,48,QtGui.QImage.Format_ARGB32)
    im.fill(c)
    px = QtGui.QPixmap.fromImage(im)
    return QtGui.QIcon(px)


class Layer(gui_base.GuiCommandSimplest):
    """GuiCommand to create a Layer object in the document."""

    def __init__(self):
        super(Layer, self).__init__(name=translate("draft","Layer"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {'Pixmap': 'Draft_Layer',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Layer", "Layer"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Layer", "Adds a layer to the document.\nObjects added to this layer can share the same visual properties such as line color, line width, and shape color.")}

    def Activated(self):
        """Execute when the command is called.

        It calls the `finish(False)` method of the active Draft command.
        """
        super(Layer, self).Activated()

        self.doc.openTransaction("Create Layer")
        Gui.addModule("Draft")
        Gui.doCommand('_layer_ = Draft.make_layer()')
        Gui.doCommand('FreeCAD.ActiveDocument.recompute()')
        self.doc.commitTransaction()


class LayerManager:

    """GuiCommand that displays a Layers manager dialog"""

    def GetResources(self):

        return {'Pixmap'  : 'Draft_LayerManager',
                'MenuText': QT_TRANSLATE_NOOP("Draft_LayerManager", "Manage layers..."),
                'ToolTip' : QT_TRANSLATE_NOOP("Draft_LayerManager", "Set/modify the different layers of this document")}

    def Activated(self):

        from PySide import QtCore, QtGui

        # store changes to be committed
        self.deleteList = []

        # create the dialog
        self.dialog = Gui.PySideUic.loadUi(":/ui/dialogLayers.ui")

        # set nice icons
        self.dialog.setWindowIcon(QtGui.QIcon(":/icons/Draft_Layer.svg"))
        self.dialog.buttonNew.setIcon(QtGui.QIcon(":/icons/document-new.svg"))
        self.dialog.buttonDelete.setIcon(QtGui.QIcon(":/icons/delete.svg"))
        self.dialog.buttonSelectAll.setIcon(QtGui.QIcon(":/icons/edit-select-all.svg"))
        self.dialog.buttonToggle.setIcon(QtGui.QIcon(":/icons/dagViewVisible.svg"))
        self.dialog.buttonIsolate.setIcon(QtGui.QIcon(":/icons/view-refresh.svg"))
        self.dialog.buttonCancel.setIcon(QtGui.QIcon(":/icons/edit_Cancel.svg"))
        self.dialog.buttonOK.setIcon(QtGui.QIcon(":/icons/edit_OK.svg"))

        # restore window geometry from stored state
        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")
        w = pref.GetInt("LayersManagerWidth",640)
        h = pref.GetInt("LayersManagerHeight",320)
        self.dialog.resize(w,h)

        # center the dialog over FreeCAD window
        mw = Gui.getMainWindow()
        self.dialog.move(mw.frameGeometry().topLeft() + mw.rect().center() - self.dialog.rect().center())

        # connect signals/slots
        self.dialog.buttonNew.clicked.connect(self.addItem)
        self.dialog.buttonDelete.clicked.connect(self.onDelete)
        self.dialog.buttonSelectAll.clicked.connect(self.dialog.tree.selectAll)
        self.dialog.buttonToggle.clicked.connect(self.onToggle)
        self.dialog.buttonCancel.clicked.connect(self.dialog.reject)
        self.dialog.buttonIsolate.clicked.connect(self.onIsolate)
        self.dialog.buttonOK.clicked.connect(self.accept)
        self.dialog.rejected.connect(self.reject)

        # set the model up
        self.model = QtGui.QStandardItemModel()
        self.dialog.tree.setModel(self.model)
        self.dialog.tree.setUniformRowHeights(True)
        self.dialog.tree.setItemDelegate(Layers_Delegate())
        self.dialog.tree.setItemsExpandable(False)
        self.dialog.tree.setRootIsDecorated(False) # removes spacing in first column
        self.dialog.tree.setSelectionMode(QtGui.QTreeView.ExtendedSelection) # allow to select many

        # fill the tree view
        self.update()

        # rock 'n roll!!!
        self.dialog.exec_()

    def accept(self):

        "when OK button is pressed"

        changed = False

        # delete layers
        for name in self.deleteList:
            if not changed:
                FreeCAD.ActiveDocument.openTransaction("Layers change")
                changed = True
            FreeCAD.ActiveDocument.removeObject(name)

        # apply changes
        for row in range(self.model.rowCount()):

            # get or create layer
            name = self.model.item(row,1).toolTip()
            obj = None
            if name:
                obj = FreeCAD.ActiveDocument.getObject(name)
            if not obj:
                if not changed:
                    FreeCAD.ActiveDocument.openTransaction("Layers change")
                    changed = True
                obj = Draft.make_layer()

            # visibility
            checked = True if self.model.item(row,0).checkState() == QtCore.Qt.Checked else False
            if checked != obj.ViewObject.Visibility:
                if not changed:
                    FreeCAD.ActiveDocument.openTransaction("Layers change")
                    changed = True
                obj.ViewObject.Visibility = checked

            # label
            label = self.model.item(row,1).text()
            if label:
                if obj.Label != label:
                    if not changed:
                        FreeCAD.ActiveDocument.openTransaction("Layers change")
                        changed = True
                    obj.Label = label

            # line width
            width = self.model.item(row,2).data(QtCore.Qt.DisplayRole)
            if width:
                if obj.ViewObject.LineWidth != width:
                    if not changed:
                        FreeCAD.ActiveDocument.openTransaction("Layers change")
                        changed = True
                    obj.ViewObject.LineWidth = width

            # draw style
            style = self.model.item(row,3).text()
            if style:
                if obj.ViewObject.DrawStyle != style:
                    if not changed:
                        FreeCAD.ActiveDocument.openTransaction("Layers change")
                        changed = True
                    obj.ViewObject.DrawStyle = style

            # line color
            color = self.model.item(row,4).data(QtCore.Qt.UserRole)
            if color:
                if obj.ViewObject.LineColor[3:] != color:
                    if not changed:
                        FreeCAD.ActiveDocument.openTransaction("Layers change")
                        changed = True
                    obj.ViewObject.LineColor = color

            # shape color
            color = self.model.item(row,5).data(QtCore.Qt.UserRole)
            if color:
                if obj.ViewObject.ShapeColor[3:] != color:
                    if not changed:
                        FreeCAD.ActiveDocument.openTransaction("Layers change")
                        changed = True
                    obj.ViewObject.ShapeColor = color

            # transparency
            transparency = self.model.item(row,6).data(QtCore.Qt.DisplayRole)
            if transparency:
                if obj.ViewObject.Transparency != transparency:
                    if not changed:
                        FreeCAD.ActiveDocument.openTransaction("Layers change")
                        changed = True
                    obj.ViewObject.Transparency = transparency

            # line print color
            color = self.model.item(row,7).data(QtCore.Qt.UserRole)
            if color:
                if not "LinePrintColor" in obj.ViewObject.PropertiesList:
                    if hasattr(obj.ViewObject.Proxy,"set_properties"):
                        obj.ViewObject.Proxy.set_properties(obj.ViewObject)
                if "LinePrintColor" in obj.ViewObject.PropertiesList:
                    if obj.ViewObject.LinePrintColor[3:] != color:
                        if not changed:
                            FreeCAD.ActiveDocument.openTransaction("Layers change")
                            changed = True
                        obj.ViewObject.LinePrintColor = color

        # recompute
        if changed:
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()

        # exit
        self.dialog.reject()

    def reject(self):

        "when Cancel button is pressed or dialog is closed"

        # save dialog size
        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")
        pref.SetInt("LayersManagerWidth",self.dialog.width())
        pref.SetInt("LayersManagerHeight",self.dialog.height())

        return True

    def update(self):

        "rebuild the model from document contents"

        self.model.clear()

        # set header
        self.model.setHorizontalHeaderLabels([translate("Draft","On"),
                                              translate("Draft","Name"),
                                              translate("Draft","Line width"),
                                              translate("Draft","Draw style"),
                                              translate("Draft","Line color"),
                                              translate("Draft","Face color"),
                                              translate("Draft","Transparency"),
                                              translate("Draft","Line print color")])
        self.dialog.tree.header().setDefaultSectionSize(72)
        self.dialog.tree.setColumnWidth(0,32) # on/off column
        self.dialog.tree.setColumnWidth(1,128) # name column

        # populate
        objs = [obj for obj in FreeCAD.ActiveDocument.Objects if Draft.getType(obj) == "Layer"]
        objs.sort(key=lambda o:o.Label)
        for obj in objs:
            self.addItem(obj)

    def addItem(self,obj=None):

        "adds a row to the model"

        from PySide import QtCore, QtGui

        # create row with default values
        onItem = QtGui.QStandardItem()
        onItem.setCheckable(True)
        onItem.setCheckState(QtCore.Qt.Checked)
        nameItem = QtGui.QStandardItem(translate("Draft","New Layer"))
        widthItem = QtGui.QStandardItem()
        widthItem.setData(self.getPref("DefaultShapeLineWidth",2,"Integer"),QtCore.Qt.DisplayRole)
        styleItem = QtGui.QStandardItem("Solid")
        lineColorItem = QtGui.QStandardItem()
        lineColorItem.setData(self.getPref("DefaultShapeLineColor",421075455),QtCore.Qt.UserRole)
        shapeColorItem = QtGui.QStandardItem()
        shapeColorItem.setData(self.getPref("DefaultShapeColor",3435973887),QtCore.Qt.UserRole)
        transparencyItem = QtGui.QStandardItem()
        transparencyItem.setData(0,QtCore.Qt.DisplayRole)
        linePrintColorItem = QtGui.QStandardItem()
        linePrintColorItem.setData(self.getPref("DefaultPrintColor",0),QtCore.Qt.UserRole)

        # populate with object data
        if obj:
            onItem.setCheckState(QtCore.Qt.Checked if obj.ViewObject.Visibility else QtCore.Qt.Unchecked)
            nameItem.setText(obj.Label)
            nameItem.setToolTip(obj.Name)
            widthItem.setData(obj.ViewObject.LineWidth,QtCore.Qt.DisplayRole)
            styleItem.setText(obj.ViewObject.DrawStyle)
            lineColorItem.setData(obj.ViewObject.LineColor[:3],QtCore.Qt.UserRole)
            shapeColorItem.setData(obj.ViewObject.ShapeColor[:3],QtCore.Qt.UserRole)
            transparencyItem.setData(obj.ViewObject.Transparency,QtCore.Qt.DisplayRole)
            if hasattr(obj.ViewObject,"LinePrintColor"):
                linePrintColorItem.setData(obj.ViewObject.LinePrintColor[:3],QtCore.Qt.UserRole)
        lineColorItem.setIcon(getColorIcon(lineColorItem.data(QtCore.Qt.UserRole)))
        shapeColorItem.setIcon(getColorIcon(shapeColorItem.data(QtCore.Qt.UserRole)))
        linePrintColorItem.setIcon(getColorIcon(linePrintColorItem.data(QtCore.Qt.UserRole)))

        # append row
        self.model.appendRow([onItem,
                              nameItem,
                              widthItem,
                              styleItem,
                              lineColorItem,
                              shapeColorItem,
                              transparencyItem,
                              linePrintColorItem])

    def getPref(self,value,default,valuetype="Unsigned"):

        "retrieves a view pref value"

        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View")
        if valuetype == "Unsigned":
            c = p.GetUnsigned(value,default)
            r = float((c>>24)&0xFF)/255.0
            g = float((c>>16)&0xFF)/255.0
            b = float((c>>8)&0xFF)/255.0
            return (r,g,b,)
        elif valuetype == "Integer":
            return p.GetInt(value,default)

    def onDelete(self):

        "delete selected rows"

        rows = []
        for index in self.dialog.tree.selectedIndexes():
            if not index.row() in rows:
                rows.append(index.row())

            # append layer name to the delete list
            if index.column() == 1:
                name = self.model.itemFromIndex(index).toolTip()
                if name:
                    if not name in self.deleteList:
                        self.deleteList.append(name)

        # delete rows starting from the lowest, to not alter row indexes while deleting
        rows.sort()
        rows.reverse()
        for row in rows:
            self.model.takeRow(row)

    def onToggle(self):

        "toggle selected layers on/off"

        from PySide import QtCore, QtGui

        state = None
        for index in self.dialog.tree.selectedIndexes():
            if index.column() == 0:
                # get state from first selected row
                if state is None:
                    if self.model.itemFromIndex(index).checkState() == QtCore.Qt.Checked:
                        state = QtCore.Qt.Unchecked
                    else:
                        state = QtCore.Qt.Checked
                self.model.itemFromIndex(index).setCheckState(state)

    def onIsolate(self):

        "isolates the selected layers (turns all the others off"

        from PySide import QtCore, QtGui

        onrows = []
        for index in self.dialog.tree.selectedIndexes():
            if not index.row() in onrows:
                onrows.append(index.row())
        for row in range(self.model.rowCount()):
            if not row in onrows:
                self.model.item(row,0).setCheckState(QtCore.Qt.Unchecked)


if FreeCAD.GuiUp:

    from PySide import QtCore, QtGui

    class Layers_Delegate(QtGui.QStyledItemDelegate):

        "model delegate"

        def __init__(self, parent=None, *args):

            QtGui.QStyledItemDelegate.__init__(self, parent, *args)
            # setEditorData() is triggered several times.
            # But we want to show the color dialog only the first time
            self.first = True

        def createEditor(self,parent,option,index):

            if index.column() == 0: # Layer on/off
                editor = QtGui.QCheckBox(parent)
            if index.column() == 1: # Layer name
                editor = QtGui.QLineEdit(parent)
            elif index.column() == 2: # Line width
                editor = QtGui.QSpinBox(parent)
                editor.setMaximum(99)
            elif index.column() == 3: # Line style
                editor = QtGui.QComboBox(parent)
                editor.addItems(["Solid","Dashed","Dotted","Dashdot"])
            elif index.column() == 4: # Line color
                editor = QtGui.QLineEdit(parent)
                self.first = True
            elif index.column() == 5: # Shape color
                editor = QtGui.QLineEdit(parent)
                self.first = True
            elif index.column() == 6: # Transparency
                editor = QtGui.QSpinBox(parent)
                editor.setMaximum(100)
            elif index.column() == 7: # Line print color
                editor = QtGui.QLineEdit(parent)
                self.first = True
            return editor

        def setEditorData(self, editor, index):

            if index.column() == 0: # Layer on/off
                editor.setChecked(index.data())
            elif index.column() == 1: # Layer name
                editor.setText(index.data())
            elif index.column() == 2: # Line width
                editor.setValue(index.data())
            elif index.column() == 3: # Line style
                editor.setCurrentIndex(["Solid","Dashed","Dotted","Dashdot"].index(index.data()))
            elif index.column() == 4: # Line color
                editor.setText(str(index.data(QtCore.Qt.UserRole)))
                if self.first:
                    c = index.data(QtCore.Qt.UserRole)
                    color = QtGui.QColorDialog.getColor(QtGui.QColor(int(c[0]*255),int(c[1]*255),int(c[2]*255)))
                    editor.setText(str(color.getRgbF()))
                    self.first = False
            elif index.column() == 5: # Shape color
                editor.setText(str(index.data(QtCore.Qt.UserRole)))
                if self.first:
                    c = index.data(QtCore.Qt.UserRole)
                    color = QtGui.QColorDialog.getColor(QtGui.QColor(int(c[0]*255),int(c[1]*255),int(c[2]*255)))
                    editor.setText(str(color.getRgbF()))
                    self.first = False
            elif index.column() == 6: # Transparency
                editor.setValue(index.data())
            elif index.column() == 7:  # Line print color
                editor.setText(str(index.data(QtCore.Qt.UserRole)))
                if self.first:
                    c = index.data(QtCore.Qt.UserRole)
                    color = QtGui.QColorDialog.getColor(QtGui.QColor(int(c[0]*255),int(c[1]*255),int(c[2]*255)))
                    editor.setText(str(color.getRgbF()))
                    self.first = False

        def setModelData(self, editor, model, index):

            if index.column() == 0: # Layer on/off
                model.setData(index,editor.isChecked())
            elif index.column() == 1: # Layer name
                model.setData(index,editor.text())
            elif index.column() == 2: # Line width
                model.setData(index,editor.value())
            elif index.column() == 3: # Line style
                model.setData(index,["Solid","Dashed","Dotted","Dashdot"][editor.currentIndex()])
            elif index.column() == 4: # Line color
                model.setData(index,eval(editor.text()),QtCore.Qt.UserRole)
                model.itemFromIndex(index).setIcon(getColorIcon(eval(editor.text())))
            elif index.column() == 5: # Shape color
                model.setData(index,eval(editor.text()),QtCore.Qt.UserRole)
                model.itemFromIndex(index).setIcon(getColorIcon(eval(editor.text())))
            elif index.column() == 6: # Transparency
                model.setData(index,editor.value())
            elif index.column() == 7: # Line prin color
                model.setData(index,eval(editor.text()),QtCore.Qt.UserRole)
                model.itemFromIndex(index).setIcon(getColorIcon(eval(editor.text())))




Gui.addCommand('Draft_Layer', Layer())
Gui.addCommand('Draft_LayerManager', LayerManager())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

from time import sleep
import unittest
import FreeCAD

from Addon import Addon

from PySide import QtCore, QtWidgets

from addonmanager_update_all_gui import UpdateAllGUI, AddonStatus


class MockUpdater(QtCore.QObject):
    success = QtCore.Signal(object)
    failure = QtCore.Signal(object)
    finished = QtCore.Signal()

    def __init__(self, addon, addons=[]):
        super().__init__()
        self.addon_to_install = addon
        self.addons = addons
        self.has_run = False
        self.emit_success = True
        self.work_function = (
            None  # Set to some kind of callable to make this function take time
        )

    def run(self):
        self.has_run = True
        if self.work_function is not None and callable(self.work_function):
            self.work_function()
        if self.emit_success:
            self.success.emit(self.addon_to_install)
        else:
            self.failure.emit(self.addon_to_install)
        self.finished.emit()


class MockUpdaterFactory:
    def __init__(self, addons):
        self.addons = addons
        self.work_function = None
        self.updater = None

    def get_updater(self, addon):
        self.updater = MockUpdater(addon, self.addons)
        self.updater.work_function = self.work_function
        return self.updater


class MockAddon:
    def __init__(self, name):
        self.display_name = name
        self.name = name
        self.macro = None

    def status(self):
        return Addon.Status.UPDATE_AVAILABLE


class CallInterceptor:
    def __init__(self):
        self.called = False
        self.args = None

    def intercept(self, *args):
        self.called = True
        self.args = args


class TestUpdateAllGui(unittest.TestCase):
    def setUp(self):
        self.addons = []
        for i in range(3):
            self.addons.append(MockAddon(f"Mock Addon {i}"))
        self.factory = MockUpdaterFactory(self.addons)
        self.test_object = UpdateAllGUI(self.addons)
        self.test_object.updater_factory = self.factory

    def tearDown(self):
        pass

    def test_run(self):
        self.factory.work_function = lambda: sleep(0.1)
        self.test_object.run()
        while self.test_object.is_running():
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 100)
        self.test_object.dialog.accept()

    def test_setup_dialog(self):
        self.test_object._setup_dialog()
        self.assertIsNotNone(
            self.test_object.dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel)
        )
        self.assertEqual(self.test_object.dialog.tableWidget.rowCount(), 3)

    def test_cancelling_installation(self):
        class Worker:
            def __init__(self):
                self.counter = 0
                self.LIMIT = 100
                self.limit_reached = False

            def run(self):
                while self.counter < self.LIMIT:
                    if QtCore.QThread.currentThread().isInterruptionRequested():
                        return
                    self.counter += 1
                    sleep(0.01)
                self.limit_reached = True

        worker = Worker()
        self.factory.work_function = worker.run
        self.test_object.run()
        cancel_timer = QtCore.QTimer()
        cancel_timer.timeout.connect(
            self.test_object.dialog.buttonBox.button(
                QtWidgets.QDialogButtonBox.Cancel
            ).click
        )
        cancel_timer.start(90)
        while self.test_object.is_running():
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 10)
        self.assertGreater(len(self.test_object.addons_with_update), 0)

    def test_add_addon_to_table(self):
        mock_addon = MockAddon("MockAddon")
        self.test_object.dialog.tableWidget.clear()
        self.test_object._add_addon_to_table(mock_addon)
        self.assertEqual(self.test_object.dialog.tableWidget.rowCount(), 1)

    def test_update_addon_status(self):
        self.test_object._setup_dialog()
        self.test_object._update_addon_status(0, AddonStatus.WAITING)
        self.assertEqual(
            self.test_object.dialog.tableWidget.item(0, 1).text(),
            AddonStatus.WAITING.ui_string(),
        )
        self.test_object._update_addon_status(0, AddonStatus.INSTALLING)
        self.assertEqual(
            self.test_object.dialog.tableWidget.item(0, 1).text(),
            AddonStatus.INSTALLING.ui_string(),
        )
        self.test_object._update_addon_status(0, AddonStatus.SUCCEEDED)
        self.assertEqual(
            self.test_object.dialog.tableWidget.item(0, 1).text(),
            AddonStatus.SUCCEEDED.ui_string(),
        )
        self.test_object._update_addon_status(0, AddonStatus.FAILED)
        self.assertEqual(
            self.test_object.dialog.tableWidget.item(0, 1).text(),
            AddonStatus.FAILED.ui_string(),
        )

    def test_process_next_update(self):
        self.test_object._setup_dialog()
        self.test_object._launch_active_installer = lambda: None
        self.test_object._process_next_update()
        self.assertEqual(
            self.test_object.dialog.tableWidget.item(0, 1).text(),
            AddonStatus.INSTALLING.ui_string(),
        )

        self.test_object._process_next_update()
        self.assertEqual(
            self.test_object.dialog.tableWidget.item(1, 1).text(),
            AddonStatus.INSTALLING.ui_string(),
        )

        self.test_object._process_next_update()
        self.assertEqual(
            self.test_object.dialog.tableWidget.item(2, 1).text(),
            AddonStatus.INSTALLING.ui_string(),
        )

        self.test_object._process_next_update()

    def test_launch_active_installer(self):
        self.test_object.active_installer = self.factory.get_updater(self.addons[0])
        self.test_object._update_succeeded = lambda _: None
        self.test_object._update_failed = lambda _: None
        self.test_object.process_next_update = lambda: None
        self.test_object._launch_active_installer()
        # The above call does not block, so spin until it has completed (basically instantly in testing)
        while self.test_object.worker_thread.isRunning():
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 100)
        self.test_object.dialog.accept()

    def test_update_succeeded(self):
        self.test_object._setup_dialog()
        self.test_object._update_succeeded(self.addons[0])
        self.assertEqual(
            self.test_object.dialog.tableWidget.item(0, 1).text(),
            AddonStatus.SUCCEEDED.ui_string(),
        )

    def test_update_failed(self):
        self.test_object._setup_dialog()
        self.test_object._update_failed(self.addons[0])
        self.assertEqual(
            self.test_object.dialog.tableWidget.item(0, 1).text(),
            AddonStatus.FAILED.ui_string(),
        )

    def test_update_finished(self):
        self.test_object._setup_dialog()
        call_interceptor = CallInterceptor()
        self.test_object.worker_thread = QtCore.QThread()
        self.test_object.worker_thread.start()
        self.test_object._process_next_update = call_interceptor.intercept
        self.test_object.active_installer = self.factory.get_updater(self.addons[0])
        self.test_object._update_finished()
        self.assertFalse(self.test_object.worker_thread.isRunning())
        self.test_object.worker_thread.quit()
        self.assertTrue(call_interceptor.called)
        self.test_object.worker_thread.wait()

    def test_finalize(self):
        self.test_object._setup_dialog()
        self.test_object.worker_thread = QtCore.QThread()
        self.test_object.worker_thread.start()
        self.test_object._finalize()
        self.assertFalse(self.test_object.worker_thread.isRunning())
        self.test_object.worker_thread.quit()
        self.test_object.worker_thread.wait()
        self.assertFalse(self.test_object.running)
        self.assertIsNotNone(
            self.test_object.dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Close)
        )
        self.assertIsNone(
            self.test_object.dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel)
        )

    def test_is_running(self):
        self.assertFalse(self.test_object.is_running())
        self.test_object.run()
        self.assertTrue(self.test_object.is_running())
        while self.test_object.is_running():
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 100)
        self.test_object.dialog.accept()



# ***************************************************************************
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the initialization code for the workbench's status bar.

The status bar is activated by `InitGui.py` when the workbench is started,
and is populated by various widgets, buttons and menus.
"""
## @package init_draft_statusbar
# \ingroup draftutils
# \brief Provides the initialization code for the workbench's status bar.

## \addtogroup draftutils
# @{
import PySide.QtCore as QtCore
import PySide.QtGui as QtGui

import FreeCAD as App
import FreeCADGui as Gui

from draftutils.init_tools import get_draft_snap_commands
from draftutils.translate import translate

#----------------------------------------------------------------------------
# SCALE WIDGET FUNCTIONS
#----------------------------------------------------------------------------

draft_scales_metrics =  ["1:1000", "1:500", "1:250", "1:200", "1:100",
                         "1:50", "1:25","1:20", "1:10", "1:5","1:2",
                         "1:1",
                         "2:1", "5:1", "10:1", "20:1",
                         translate("draft", "Custom"),
                        ]

draft_scales_arch_imperial =  ["1/16in=1ft", "3/32in=1ft", "1/8in=1ft",
                               "3/16in=1ft", "1/4in=1ft","3/8in=1ft",
                               "1/2in=1ft", "3/4in=1ft", "1in=1ft",
                               "1.5in=1ft", "3in=1ft",
                               translate("draft", "Custom"),
                              ]

draft_scales_eng_imperial =  ["1in=10ft", "1in=20ft", "1in=30ft",
                              "1in=40ft", "1in=50ft", "1in=60ft",
                              "1in=70ft", "1in=80ft", "1in=90ft",
                              "1in=100ft",
                              translate("draft", "Custom"),
                             ]

def get_scales(unit_system = 0):
    """
    returns the list of preset scales according to unit system.

    Parameters:
    unit_system =   0 : default from user preferences
                    1 : metrics
                    2 : imperial architectural
                    3 : imperial engineering
    """

    if unit_system == 0:
        param = App.ParamGet("User parameter:BaseApp/Preferences/Units")
        scale_units_system = param.GetInt("UserSchema", 0)
        if scale_units_system in [0, 1, 4, 6]:
            return draft_scales_metrics
        elif scale_units_system in [2, 3, 5]:
            return draft_scales_arch_imperial
        elif scale_units_system in [7]:
            return draft_scales_eng_imperial
    elif unit_system == 1:
        return draft_scales_metrics
    elif unit_system == 2:
        return draft_scales_arch_imperial
    elif unit_system == 3:
        return draft_scales_eng_imperial


def scale_to_label(scale):
    """
    transform a float number into a 1:X or X:1 scale and return it as label
    """
    f = round(scale, 2)
    if f == 1.0:
        return "1:1"
    elif f > 1.0:
        f = f.as_integer_ratio()
        if f[1] == 1:
            return str(f[0]) + ":1"
        else:
            return str(scale)
    else:
        f = round(1/scale, 2)
        f = f.as_integer_ratio()
        if f[1] == 1:
            return "1:" + str(f[0])
        else:
            return str(scale)

def label_to_scale(label):
    """
    transform a scale string into scale factor as float
    """
    try :
        scale = float(label)
        return scale
    except Exception:
        if ":" in label:
            f = label.split(":")
        elif "=" in label:
            f = label.split("=")
        else:
            return
        if len(f) == 2:
            try:
                num = App.Units.Quantity(f[0]).Value
                den = App.Units.Quantity(f[1]).Value
                scale = num/den
                return scale
            except Exception:
                err = translate("draft",
                                "Unable to convert input into a scale factor")
                App.Console.PrintWarning(err)
                return None

def _set_scale(action):
    """
    triggered by scale pushbutton, set DraftAnnotationScale in preferences
    """
    # set the label of the scale button
    param = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")

    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    scale_widget = sb.findChild(QtGui.QToolBar,"draft_scale_widget")

    if action.text() == translate("draft", "Custom"):
        title_text = translate("draft", "Set custom scale")
        dialog_text = translate("draft",
                                "Set custom annotation scale in format x:x, x=x")
        custom_scale = QtGui.QInputDialog.getText(None, title_text, dialog_text)
        if custom_scale[1]:
            print(custom_scale[0])
            scale = label_to_scale(custom_scale[0])
            if scale is None:
                return
            param.SetFloat("DraftAnnotationScale", scale)
            cs = scale_to_label(scale)
            scale_widget.scaleLabel.setText(cs)
    else:
        text_scale = action.text()
        scale_widget.scaleLabel.setText(text_scale)
        scale = label_to_scale(text_scale)
        param.SetFloat("DraftAnnotationScale", scale)

#----------------------------------------------------------------------------
# MAIN DRAFT STATUSBAR FUNCTIONS
#----------------------------------------------------------------------------

def init_draft_statusbar_scale():
    """
    this function initializes draft statusbar scale widget
    """
    param = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")

    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    scale_widget = QtGui.QToolBar()
    scale_widget.setObjectName("draft_scale_widget")

    # get scales list according to system units
    draft_scales = get_scales()

    # get draft annotation scale
    draft_annotation_scale = param.GetFloat("DraftAnnotationScale", 1.0)

    # initializes scale widget
    scale_widget.draft_scales = draft_scales
    scaleLabel = QtGui.QPushButton("Scale")
    scaleLabel.setObjectName("ScaleLabel")
    scaleLabel.setFlat(True)
    menu = QtGui.QMenu(scaleLabel)
    gUnits = QtGui.QActionGroup(menu)
    for u in draft_scales:
        a = QtGui.QAction(gUnits)
        a.setText(u)
        menu.addAction(a)
    scaleLabel.setMenu(menu)
    gUnits.triggered.connect(_set_scale)
    scale_label = scale_to_label(draft_annotation_scale)
    scaleLabel.setText(scale_label)
    scaleLabel.setToolTip(translate("draft",
                                    "Set the scale used by draft annotation tools"))
    scale_widget.addWidget(scaleLabel)
    scale_widget.scaleLabel = scaleLabel

    # add scale widget to the statusbar
    sb.insertPermanentWidget(3, scale_widget)
    scale_widget.show()


def init_draft_statusbar_snap():
    """
    this function initializes draft statusbar snap widget
    """

    def _spacer():
        """
        empty label instead of snap_widget.addSeparator()
        """
        label = QtGui.QLabel()
        label.setFixedWidth(2)
        return label

    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    # check if the toolbar is available, without it the required actions
    # may be missing:
    if mw.findChild(QtGui.QToolBar, "Draft snap") is None:
        return

    # snap widget:
    snap_widget = QtGui.QToolBar()
    snap_widget.setObjectName("draft_snap_widget")
    snap_widget.setOrientation(QtCore.Qt.Orientation.Horizontal)
    snap_widget.setIconSize(QtCore.QSize(16, 16))
    sb.insertPermanentWidget(2, snap_widget)

    # grid button:
    snap_widget.addAction(Gui.Command.get("Draft_ToggleGrid").getAction()[0])

    snap_widget.addWidget(_spacer())

    # lock button:
    snap_widget.addAction(Gui.Command.get("Draft_Snap_Lock").getAction()[0])
    snap_action = snap_widget.children()[-1]
    snap_action.setFixedWidth(40) # Widen the button.

    snap_widget.addWidget(_spacer())

    # dimension button:
    snap_widget.addAction(Gui.Command.get("Draft_Snap_Dimensions").getAction()[0])

    snap_widget.addWidget(_spacer())

    # ortho button:
    snap_widget.addAction(Gui.Command.get("Draft_Snap_Ortho").getAction()[0])

    snap_widget.addWidget(_spacer())

    # working plane button:
    snap_widget.addAction(Gui.Command.get("Draft_Snap_WorkingPlane").getAction()[0])

    # menu for lock button:
    for cmd in get_draft_snap_commands():
        if cmd not in ["Separator",
                       "Draft_ToggleGrid",
                       "Draft_Snap_Lock", # Is automatically added to the menu anyway.
                       "Draft_Snap_Dimensions",
                       "Draft_Snap_Ortho",
                       "Draft_Snap_WorkingPlane"]:
            snap_action.addAction(Gui.Command.get(cmd).getAction()[0])


def show_draft_statusbar():
    """
    shows draft statusbar if present or initializes it
    """
    params = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")

    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    if params.GetBool("DisplayStatusbarScaleWidget", True):
        scale_widget = sb.findChild(QtGui.QToolBar, "draft_scale_widget")
        if scale_widget:
            scale_widget.show()
        else:
            scale_widget = mw.findChild(QtGui.QToolBar, "draft_scale_widget")
            if scale_widget:
                sb.insertPermanentWidget(3, scale_widget)
                scale_widget.show()
            else:
                t = QtCore.QTimer()
                t.singleShot(500, init_draft_statusbar_scale)

    if params.GetBool("DisplayStatusbarSnapWidget", True):
        snap_widget = sb.findChild(QtGui.QToolBar, "draft_snap_widget")
        if snap_widget:
            snap_widget.setOrientation(QtCore.Qt.Orientation.Horizontal)
            snap_widget.show()
        else:
            snap_widget = mw.findChild(QtGui.QToolBar, "draft_snap_widget")
            if snap_widget:
                sb.insertPermanentWidget(2, snap_widget)
                snap_widget.setOrientation(QtCore.Qt.Orientation.Horizontal)
                snap_widget.show()
            else:
                t = QtCore.QTimer()
                t.singleShot(500, init_draft_statusbar_snap)


def hide_draft_statusbar():
    """
    hides draft statusbar if present
    """
    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    # hide scale widget
    scale_widget = sb.findChild(QtGui.QToolBar, "draft_scale_widget")
    if scale_widget is None:
        # when switching workbenches, the toolbar sometimes "jumps"
        # out of the status bar to any other dock area...
        scale_widget = mw.findChild(QtGui.QToolBar, "draft_scale_widget")
    if scale_widget:
        scale_widget.hide()
        # prevent the widget from showing up as a blank item in the toolbar
        # area context menu after switching to a different workbench:
        scale_widget.toggleViewAction().setVisible(False)

    # hide snap widget
    snap_widget = sb.findChild(QtGui.QToolBar,"draft_snap_widget")
    if snap_widget is None:
        # when switching workbenches, the toolbar sometimes "jumps"
        # out of the status bar to any other dock area...
        snap_widget = mw.findChild(QtGui.QToolBar,"draft_snap_widget")
    if snap_widget:
        snap_widget.hide()
        # prevent the widget from showing up as a blank item in the toolbar
        # area context menu after switching to a different workbench:
        snap_widget.toggleViewAction().setVisible(False)

## @}



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtCore, QtGui
import FreeCADGui
import Path
import Path.Base.Gui.PropertyEditor as PathPropertyEditor
import Path.Base.Gui.Util as PathGuiUtil
import Path.Base.Util as PathUtil
import os
import re


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class _Delegate(QtGui.QStyledItemDelegate):
    """Handles the creation of an appropriate editing widget for a given property."""

    ObjectRole = QtCore.Qt.UserRole + 1
    PropertyRole = QtCore.Qt.UserRole + 2
    EditorRole = QtCore.Qt.UserRole + 3

    def createEditor(self, parent, option, index):
        editor = index.data(self.EditorRole)
        if editor is None:
            obj = index.data(self.ObjectRole)
            prp = index.data(self.PropertyRole)
            editor = PathPropertyEditor.Editor(obj, prp)
            index.model().setData(index, editor, self.EditorRole)
        return editor.widget(parent)

    def setEditorData(self, widget, index):
        # called to update the widget with the current data
        index.data(self.EditorRole).setEditorData(widget)

    def setModelData(self, widget, model, index):
        # called to update the model with the data from the widget
        editor = index.data(self.EditorRole)
        editor.setModelData(widget)
        index.model().setData(
            index,
            PathUtil.getPropertyValueString(editor.obj, editor.prop),
            QtCore.Qt.DisplayRole,
        )


class ToolBitEditor(object):
    """UI and controller for editing a ToolBit.
    The controller embeds the UI to the parentWidget which has to have a
    layout attached to it.
    """

    def __init__(self, tool, parentWidget=None, loadBitBody=True):
        Path.Log.track()
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/ToolBitEditor.ui")

        if parentWidget:
            self.form.setParent(parentWidget)
            parentWidget.layout().addWidget(self.form)

        self.tool = tool
        self.loadbitbody = loadBitBody
        if not tool.BitShape:
            self.tool.BitShape = "endmill.fcstd"

        if self.loadbitbody:
            self.tool.Proxy.loadBitBody(self.tool)

        # remove example widgets
        layout = self.form.bitParams.layout()
        for i in range(layout.rowCount() - 1, -1, -1):
            layout.removeRow(i)
        # used to track property widgets and editors
        self.widgets = []

        self.setupTool(self.tool)
        self.setupAttributes(self.tool)

    def setupTool(self, tool):
        Path.Log.track()
        # Can't delete and add fields to the form because of dangling references in case of
        # a focus change. see https://forum.freecad.org/viewtopic.php?f=10&t=52246#p458583
        # Instead we keep widgets once created and use them for new properties, and hide all
        # which aren't being needed anymore.

        def labelText(name):
            return re.sub("([A-Z][a-z]+)", r" \1", re.sub("([A-Z]+)", r" \1", name))

        layout = self.form.bitParams.layout()
        ui = FreeCADGui.UiLoader()

        # for all properties either assign them to existing labels and editors
        # or create additional ones for them if not enough have already been
        # created.
        usedRows = 0
        for nr, name in enumerate(tool.Proxy.toolShapeProperties(tool)):
            if nr < len(self.widgets):
                Path.Log.debug("re-use row: {} [{}]".format(nr, name))
                label, qsb, editor = self.widgets[nr]
                label.setText(labelText(name))
                editor.attachTo(tool, name)
                label.show()
                qsb.show()
            else:
                qsb = ui.createWidget("Gui::QuantitySpinBox")
                editor = PathGuiUtil.QuantitySpinBox(qsb, tool, name)
                label = QtGui.QLabel(labelText(name))
                self.widgets.append((label, qsb, editor))
                Path.Log.debug("create row: {} [{}]  {}".format(nr, name, type(qsb)))
                if hasattr(qsb, "editingFinished"):
                    qsb.editingFinished.connect(self.updateTool)

            if nr >= layout.rowCount():
                layout.addRow(label, qsb)
            usedRows = usedRows + 1

        # hide all rows which aren't being used
        Path.Log.track(usedRows, len(self.widgets))
        for i in range(usedRows, len(self.widgets)):
            label, qsb, editor = self.widgets[i]
            label.hide()
            qsb.hide()
            editor.attachTo(None)
            Path.Log.debug("  hide row: {}".format(i))

        img = tool.Proxy.getBitThumbnail(tool)
        if img:
            self.form.image.setPixmap(QtGui.QPixmap(QtGui.QImage.fromData(img)))
        else:
            self.form.image.setPixmap(QtGui.QPixmap())

    def setupAttributes(self, tool):
        Path.Log.track()

        setup = True
        if not hasattr(self, "delegate"):
            self.delegate = _Delegate(self.form.attrTree)
            self.model = QtGui.QStandardItemModel(self.form.attrTree)
            self.model.setHorizontalHeaderLabels(["Property", "Value"])
        else:
            self.model.removeRows(0, self.model.rowCount())
            setup = False

        attributes = tool.Proxy.toolGroupsAndProperties(tool, False)
        for name in attributes:
            group = QtGui.QStandardItem()
            group.setData(name, QtCore.Qt.EditRole)
            group.setEditable(False)
            for prop in attributes[name]:
                label = QtGui.QStandardItem()
                label.setData(prop, QtCore.Qt.EditRole)
                label.setEditable(False)

                value = QtGui.QStandardItem()
                value.setData(
                    PathUtil.getPropertyValueString(tool, prop), QtCore.Qt.DisplayRole
                )
                value.setData(tool, _Delegate.ObjectRole)
                value.setData(prop, _Delegate.PropertyRole)

                group.appendRow([label, value])
            self.model.appendRow(group)

        if setup:
            self.form.attrTree.setModel(self.model)
            self.form.attrTree.setItemDelegateForColumn(1, self.delegate)
        self.form.attrTree.expandAll()
        self.form.attrTree.resizeColumnToContents(0)
        self.form.attrTree.resizeColumnToContents(1)
        # self.form.attrTree.collapseAll()

    def accept(self):
        Path.Log.track()
        self.refresh()
        self.tool.Proxy.unloadBitBody(self.tool)

    def reject(self):
        Path.Log.track()
        self.tool.Proxy.unloadBitBody(self.tool)

    def updateUI(self):
        Path.Log.track()
        self.form.toolName.setText(self.tool.Label)
        self.form.shapePath.setText(self.tool.BitShape)

        for lbl, qsb, editor in self.widgets:
            editor.updateSpinBox()

    def _updateBitShape(self, shapePath):
        # Only need to go through this exercise if the shape actually changed.
        if self.tool.BitShape != shapePath:
            # Before setting a new  bitshape we need to make sure that none of
            # editors fires an event and tries to access its old property, which
            # might not exist anymore.
            for lbl, qsb, editor in self.widgets:
                editor.attachTo(self.tool, "File")
            self.tool.BitShape = shapePath
            self.setupTool(self.tool)
            self.form.toolName.setText(self.tool.Label)
            if self.tool.BitBody and self.tool.BitBody.ViewObject:
                if not self.tool.BitBody.ViewObject.Visibility:
                    self.tool.BitBody.ViewObject.Visibility = True
            self.setupAttributes(self.tool)
            return True
        return False

    def updateShape(self):
        Path.Log.track()
        shapePath = str(self.form.shapePath.text())
        # Only need to go through this exercise if the shape actually changed.
        if self._updateBitShape(shapePath):
            for lbl, qsb, editor in self.widgets:
                editor.updateSpinBox()

    def updateTool(self):
        Path.Log.track()

        label = str(self.form.toolName.text())
        shape = str(self.form.shapePath.text())
        if self.tool.Label != label:
            self.tool.Label = label
        self._updateBitShape(shape)

        for lbl, qsb, editor in self.widgets:
            editor.updateProperty()

        self.tool.Proxy._updateBitShape(self.tool)

    def refresh(self):
        Path.Log.track()
        self.form.blockSignals(True)
        self.updateTool()
        self.updateUI()
        self.form.blockSignals(False)

    def selectShape(self):
        Path.Log.track()
        path = self.tool.BitShape
        if not path:
            path = Path.Preferences.lastPathToolShape()
        foo = QtGui.QFileDialog.getOpenFileName(
            self.form, "Path - Tool Shape", path, "*.fcstd"
        )
        if foo and foo[0]:
            Path.Preferences.setLastPathToolShape(os.path.dirname(foo[0]))
            self.form.shapePath.setText(foo[0])
            self.updateShape()

    def setupUI(self):
        Path.Log.track()
        self.updateUI()

        self.form.toolName.editingFinished.connect(self.refresh)
        self.form.shapePath.editingFinished.connect(self.updateShape)
        self.form.shapeSet.clicked.connect(self.selectShape)



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2022 Larry Woestman <LarryWoestman2@gmail.com>          *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""
These are common functions and classes for creating custom post processors.
"""

from PySide import QtCore, QtGui

import FreeCAD

import Path
import Part

from Path.Base.MachineState import MachineState

translate = FreeCAD.Qt.translate
FreeCADGui = None
if FreeCAD.GuiUp:
    import FreeCADGui


class GCodeHighlighter(QtGui.QSyntaxHighlighter):
    def __init__(self, parent=None):
        super(GCodeHighlighter, self).__init__(parent)

        keywordFormat = QtGui.QTextCharFormat()
        keywordFormat.setForeground(QtCore.Qt.cyan)
        keywordFormat.setFontWeight(QtGui.QFont.Bold)
        keywordPatterns = ["\\bG[0-9]+\\b", "\\bM[0-9]+\\b"]

        self.highlightingRules = [
            (QtCore.QRegularExpression(pattern), keywordFormat)
            for pattern in keywordPatterns
        ]

        speedFormat = QtGui.QTextCharFormat()
        speedFormat.setFontWeight(QtGui.QFont.Bold)
        speedFormat.setForeground(QtCore.Qt.green)
        self.highlightingRules.append(
            (QtCore.QRegularExpression("\\bF[0-9\\.]+\\b"), speedFormat)
        )

    def highlightBlock(self, text):
        for pattern, hlFormat in self.highlightingRules:
            expression = QtCore.QRegularExpression(pattern)
            index = expression.match(text)
            while index.hasMatch():
                length = index.capturedLength()
                self.setFormat(index.capturedStart(), length, hlFormat)
                index = expression.match(text, index.capturedStart() + length)


class GCodeEditorDialog(QtGui.QDialog):
    def __init__(self, parent=None):
        if parent is None:
            parent = FreeCADGui.getMainWindow()
        QtGui.QDialog.__init__(self, parent)

        layout = QtGui.QVBoxLayout(self)

        # nice text editor widget for editing the gcode
        self.editor = QtGui.QTextEdit()
        font = QtGui.QFont()
        font.setFamily("Courier")
        font.setFixedPitch(True)
        font.setPointSize(10)
        self.editor.setFont(font)
        self.editor.setText("G01 X55 Y4.5 F300.0")
        layout.addWidget(self.editor)

        # OK and Cancel buttons
        self.buttons = QtGui.QDialogButtonBox(
            QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel,
            QtCore.Qt.Horizontal,
            self,
        )
        layout.addWidget(self.buttons)

        # restore placement and size
        self.paramKey = "User parameter:BaseApp/Values/Mod/Path/GCodeEditor/"
        params = FreeCAD.ParamGet(self.paramKey)
        posX = params.GetInt("posX")
        posY = params.GetInt("posY")
        if posX > 0 and posY > 0:
            self.move(posX, posY)
        width = params.GetInt("width")
        height = params.GetInt("height")
        if width > 0 and height > 0:
            self.resize(width, height)

        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

    def done(self, *args, **kwargs):
        params = FreeCAD.ParamGet(self.paramKey)
        params.SetInt("posX", self.x())
        params.SetInt("posY", self.y())
        params.SetInt("width", self.size().width())
        params.SetInt("height", self.size().height())
        return QtGui.QDialog.done(self, *args, **kwargs)


def stringsplit(commandline):
    returndict = {
        "command": None,
        "X": None,
        "Y": None,
        "Z": None,
        "A": None,
        "B": None,
        "F": None,
        "T": None,
        "S": None,
        "I": None,
        "J": None,
        "K": None,
        "txt": None,
    }
    wordlist = [a.strip() for a in commandline.split(" ")]
    if wordlist[0][0] == "(":
        returndict["command"] = "message"
        returndict["txt"] = wordlist[0]
    else:
        returndict["command"] = wordlist[0]
    for word in wordlist[1:]:
        returndict[word[0]] = word[1:]

    return returndict


def fmt(num, dec, units):
    """Use to format axis moves, feedrate, etc for decimal places and units."""
    if units == "G21":  # metric
        fnum = "%.*f" % (dec, num)
    else:  # inch
        fnum = "%.*f" % (dec, num / 25.4)  # since FreeCAD uses metric units internally
    return fnum


def editor(gcode):
    """Pops up a handy little editor to look at the code output."""
    prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Path")
    # default Max Highlighter Size = 512 Ko
    defaultMHS = 512 * 1024
    mhs = prefs.GetUnsigned("inspecteditorMaxHighlighterSize", defaultMHS)

    dia = GCodeEditorDialog()
    dia.editor.setText(gcode)
    gcodeSize = len(dia.editor.toPlainText())
    if gcodeSize <= mhs:
        # because of poor performance, syntax highlighting is
        # limited to mhs octets (default 512 KB).
        # It seems than the response time curve has an inflexion near 500 KB
        # beyond 500 KB, the response time increases exponentially.
        dia.highlighter = GCodeHighlighter(dia.editor.document())
    else:
        FreeCAD.Console.PrintMessage(
            translate(
                "Path",
                "GCode size too big ({} o), disabling syntax highlighter.".format(
                    gcodeSize
                ),
            )
        )
    result = dia.exec_()
    if result:  # If user selected 'OK' get modified G Code
        final = dia.editor.toPlainText()
    else:
        final = gcode
    return final


def fcoms(string, commentsym):
    """Filter and rebuild comments with user preferred comment symbol."""
    if len(commentsym) == 1:
        s1 = string.replace("(", commentsym)
        comment = s1.replace(")", "")
    else:
        return string
    return comment


def splitArcs(path):
    """Filter a path object and replace all G2/G3 moves with discrete G1 moves.

    Returns a Path object.
    """
    prefGrp = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Path")
    deflection = prefGrp.GetFloat("LibAreaCurveAccuarcy", 0.01)

    results = []
    if not isinstance(path, Path.Path):
        raise TypeError("path must be a Path object")

    machine = MachineState()
    for command in path.Commands:

        if command.Name not in Path.Geom.CmdMoveArc:
            machine.addCommand(command)
            results.append(command)
            continue

        edge = Path.Geom.edgeForCmd(command, machine.getPosition())
        pts = edge.discretize(Deflection=deflection)
        edges = [Part.makeLine(v1, v2) for v1, v2 in zip(pts, pts[1:])]
        for edge in edges:
            results.extend(Path.Geom.cmdsForEdge(edge))

        machine.addCommand(command)

    return Path.Path(results)



# ***************************************************************************
# *   Copyright (c) 2023 edi <edi271@a1.net>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the TechDraw AxoLengthDimension GuiCommand."""

__title__ = "TechDrawTools.CommandAxoLengthDimension"
__author__ = "edi"
__url__ = "https://www.freecad.org"
__version__ = "00.01"
__date__ = "2023/02/01"

from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui

import TechDrawTools

import TechDraw
from math import degrees

class CommandAxoLengthDimension:
    """Creates a 3D length dimension."""

    def __init__(self):
        """Initialize variables for the command that must exist at all times."""
        pass

    def GetResources(self):
        """Return a dictionary with data that will be used by the button or menu item."""
        return {'Pixmap': 'actions/TechDraw_AxoLengthDimension.svg',
                'Accel': "",
                'MenuText': QT_TRANSLATE_NOOP("TechDraw_AxoLengthDimension", "Axonometric length dimension"),
                'ToolTip': QT_TRANSLATE_NOOP("TechDraw_AxoLengthDimension", "Create an axonometric length dimension<br>\
                - select first edge to define direction and length of the dimension line<br>\
                - select second edge to define the direction of the extension lines<br>\
                - optional: select two more vertexes which define the measurement instead of the length<br>\
                  of the first selected edge")}

    def Activated(self):
        """Run the following code when the command is activated (button press)."""
        if  self.selectionTest():
            (edges,vertexes) = self.selectionTest()
            view = Gui.Selection.getSelection()[0]
            StartPt, EndPt = edges[1].Vertexes[0].Point, edges[1].Vertexes[1].Point
            extLineVec = EndPt.sub(StartPt)
            StartPt, EndPt = edges[0].Vertexes[0].Point, edges[0].Vertexes[1].Point
            dimLineVec = EndPt.sub(StartPt)
            xAxis = App.Vector(1,0,0)
            extAngle = degrees(extLineVec.getAngle(xAxis))
            lineAngle = degrees(dimLineVec.getAngle(xAxis))
            if extLineVec.y < 0.0:
                extAngle = 180-extAngle
            if dimLineVec.y < 0.0:
               lineAngle = 180-lineAngle
            if abs(extAngle-lineAngle)>0.1:
                distanceDim=TechDraw.makeDistanceDim(view,'Distance',vertexes[0].Point,vertexes[1].Point)
                distanceDim.AngleOverride = True
                distanceDim.LineAngle = lineAngle
                distanceDim.ExtensionAngle = extAngle
                distanceDim.X = (vertexes[0].Point.x+vertexes[1].Point.x)*view.Scale/2
                distanceDim.Y = (vertexes[0].Point.y+vertexes[1].Point.y)*view.Scale/2
                distanceDim.recompute()
                view.requestPaint()
            Gui.Selection.clearSelection()

    def IsActive(self):
        """Return True when the command should be active or False when it should be disabled (greyed)."""
        if App.ActiveDocument:
            return TechDrawTools.TDToolsUtil.havePage() and TechDrawTools.TDToolsUtil.haveView()
        else:
            return False

    def selectionTest(self):
        '''test correct selection'''
        if not Gui.Selection.getSelection():
            return False
        view = Gui.Selection.getSelection()[0]

        if not Gui.Selection.getSelectionEx():
            return False
        objectList = Gui.Selection.getSelectionEx()[0].SubElementNames

        if not len(objectList)>=2:
            return False

        edges = []
        vertexes = []
        for objectString in objectList:
            if objectString[0:4] == 'Edge':
                edges.append(view.getEdgeBySelection(objectString))
            elif objectString[0:6] == 'Vertex':
                vertexes.append(view.getVertexBySelection(objectString))
        if not len(edges) >= 2:
            return False

        if len(vertexes)<2:
            vertexes = []
            vertexes.append(edges[0].Vertexes[0])
            vertexes.append(edges[0].Vertexes[1])
        return(edges,vertexes)


#
# The command must be "registered" with a unique name by calling its class.
Gui.addCommand('TechDraw_AxoLengthDimension', CommandAxoLengthDimension())




#***************************************************************************
#*   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

import os
import codecs
import ntpath
# import numpy as np

import FreeCAD
import Arch
import Draft
import DraftGeomUtils
import Mesh
import MeshPart
import Part

if FreeCAD.GuiUp:
    from draftutils.translate import translate
else:
    # \cond
    def translate(context,text):
        return text
    # \endcond

## @package importOBJ
#  \ingroup ARCH
#  \brief OBJ file format exporter
#
#  This module provides tools to import & export OBJ files.
#  It is an alternative to the standard Mesh OBJ exporter
#  and supports exporting faces with more than 3 vertices
#  and supports object colors / materials

p = Draft.precision()

if open.__module__ in ['__builtin__','io']:
    pythonopen = open

def findVert(aVertex,aList):
    "finds aVertex in aList, returns index"
    for i in range(len(aList)):
        if round(aVertex.X,p) == round(aList[i].X,p):
            if round(aVertex.Y,p) == round(aList[i].Y,p):
                if round(aVertex.Z,p) == round(aList[i].Z,p):
                    return i
    return None

def getIndices(obj,shape,offsetv,offsetvn):
    "returns a list with 2 lists: vertices and face indexes, offset with the given amount"
    vlist = []
    vnlist = []
    elist = []
    flist = []
    hascurve = False
    mesh = None

    if isinstance(shape,Part.Shape):
        for e in shape.Edges:
            try:
                if not isinstance(e.Curve,Part.Line):
                    hascurve = True
            except Exception: # unimplemented curve type
                hascurve = True
            if hascurve:
                param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Mesh")
                tol = param.GetFloat("MaxDeviationExport",0.1)
                mesh = Mesh.Mesh()
                mesh.addFacets(shape.getFaces(tol))
                FreeCAD.Console.PrintWarning(translate("Arch","Found a shape containing curves, triangulating")+"\n")
                break
    elif isinstance(shape,Mesh.Mesh):
        mesh = shape

    if mesh:
        for v in mesh.Topology[0]:
            vlist.append(" "+str(round(v[0],p))+" "+str(round(v[1],p))+" "+str(round(v[2],p)))

        for vn in mesh.Facets:
            vnlist.append(" "+str(round(vn.Normal[0],p))+" "+str(round(vn.Normal[1],p))+" "+str(round(vn.Normal[2],p)))

        for i, vn in enumerate(mesh.Topology[1]):
            flist.append(" "+str(vn[0]+offsetv)+"//"+str(i+offsetvn)+" "+str(vn[1]+offsetv)+"//"+str(i+offsetvn)+" "+str(vn[2]+offsetv)+"//"+str(i+offsetvn)+" ")
    else:
        for v in shape.Vertexes:
            vlist.append(" "+str(round(v.X,p))+" "+str(round(v.Y,p))+" "+str(round(v.Z,p)))
        if not shape.Faces:
            for e in shape.Edges:
                if DraftGeomUtils.geomType(e) == "Line":
                    ei = " " + str(findVert(e.Vertexes[0],shape.Vertexes) + offsetv)
                    ei += " " + str(findVert(e.Vertexes[-1],shape.Vertexes) + offsetv)
                    elist.append(ei)
        for f in shape.Faces:
            if len(f.Wires) > 1:
                # if we have holes, we triangulate
                tris = f.tessellate(1)
                for fdata in tris[1]:
                    fi = ""
                    for vi in fdata:
                        vdata = Part.Vertex(tris[0][vi])
                        fi += " " + str(findVert(vdata,shape.Vertexes) + offsetv)
                    flist.append(fi)
            else:
                fi = ""
                edges = f.OuterWire.OrderedEdges
                # Avoid flipped normals:
                if f.Orientation == "Reversed":
                    edges.reverse()
                for e in edges:
                    v = e.Vertexes[0 if e.Orientation == "Forward" else 1]
                    ind = findVert(v,shape.Vertexes)
                    if ind is None:
                        return None,None,None,None
                    fi += " " + str(ind + offsetv)
                flist.append(fi)

    return vlist,vnlist,elist,flist


def export(exportList,filename,colors=None):

    """export(exportList,filename,colors=None):
    Called when freecad exports a file. exportList is a list
    of objects, filename is the .obj file to export (a .mtl
    file with same name will also be created together), and
    optionally colors can be a dict containing ["objectName:colorTuple"]
    pairs for use in non-GUI mode."""

    outfile = codecs.open(filename,"wb",encoding="utf8")
    ver = FreeCAD.Version()
    outfile.write("# FreeCAD v" + ver[0] + "." + ver[1] + " build" + ver[2] + " Arch module\n")
    outfile.write("# http://www.freecad.org\n")
    offsetv = 1
    offsetvn = 1
    objectslist = Draft.get_group_contents(exportList, walls=True,
                                           addgroups=True)
    objectslist = Arch.pruneIncluded(objectslist)
    filenamemtl = filename[:-4] + ".mtl"
    materials = []
    outfile.write("mtllib " + os.path.basename(filenamemtl) + "\n")

    for obj in objectslist:
        if FreeCAD.GuiUp:
            visible = obj.Visibility
        else:
            visible = True
        if not visible:
            continue

        if not (obj.isDerivedFrom("Part::Feature")
                or obj.isDerivedFrom("Mesh::Feature")
                or obj.isDerivedFrom("App::Link")):
            continue

        if obj.isDerivedFrom("App::Link"):
            # Get global placement from a Link:
            obj_place = obj.Placement
            parents = obj.Parents
            if parents:
                doc = obj.Document
                parents = [parents[0][0]] + [doc.getObject(name) for name in parents[0][1].split(".")[:-2]]
                parents.reverse()  # child-parent-grandparent order.
                for parent in parents:
                    obj_place = parent.Placement.multiply(obj_place)
        else:
            obj_place = obj.getGlobalPlacement()

        hires = None
        if FreeCAD.GuiUp:
            if obj.ViewObject.DisplayMode == "HiRes":
                # check if high-resolution object is available
                if hasattr(obj,"HiRes") and obj.HiRes:
                    if obj.HiRes.isDerivedFrom("Mesh::Feature"):
                        hires = obj.HiRes.Mesh.copy()
                    else:
                        hires = obj.HiRes.Shape.copy(False, True)
                    hires.Placement = obj_place.multiply(hires.Placement)
                if not hires \
                        and hasattr(obj,"CloneOf") \
                        and obj.CloneOf \
                        and hasattr(obj.CloneOf,"HiRes") \
                        and obj.CloneOf.HiRes:
                    if obj.CloneOf.HiRes.isDerivedFrom("Mesh::Feature"):
                        hires = obj.CloneOf.HiRes.Mesh.copy()
                    else:
                        hires = obj.CloneOf.HiRes.Shape.copy(False, True)
                    hires.Placement = obj_place.multiply(obj.CloneOf.Placement).multiply(hires.Placement)

        if hires:
            vlist,vnlist,elist,flist = getIndices(obj,hires,offsetv,offsetvn)
        elif hasattr(obj,"Shape") and obj.Shape:
            shape = obj.Shape.copy(False, True)
            shape.Placement = obj_place
            vlist,vnlist,elist,flist = getIndices(obj,shape,offsetv,offsetvn)
        elif hasattr(obj,"Mesh") and obj.Mesh:
            mesh = obj.Mesh.copy()
            mesh.Placement = obj_place
            vlist,vnlist, elist,flist = getIndices(obj,mesh,offsetv,offsetvn)

        if vlist is None:
            FreeCAD.Console.PrintError("Unable to export object "+obj.Label+". Skipping.\n")
        else:
            offsetv += len(vlist)
            offsetvn += len(vnlist)
            outfile.write("o " + obj.Label + "\n")

            # write material
            m = False
            if hasattr(obj,"Material"):
                if obj.Material:
                    if hasattr(obj.Material,"Material"):
                        outfile.write("usemtl " + obj.Material.Name + "\n")
                        materials.append(obj.Material)
                        m = True
            if not m:
                if colors:
                    if obj.Name in colors:
                        color = colors[obj.Name]
                        if color:
                            if isinstance(color[0],tuple):
                                # this is a diffusecolor. For now, use the first color - #TODO: Support per-face colors
                                color = color[0]
                            #print("found color for obj",obj.Name,":",color)
                            mn = Draft.getrgb(color,testbw=False)[1:]
                            outfile.write("usemtl color_" + mn + "\n")
                            materials.append(("color_" + mn,color,0))
                elif FreeCAD.GuiUp:
                    if hasattr(obj.ViewObject,"ShapeColor") and hasattr(obj.ViewObject,"Transparency"):
                        mn = Draft.getrgb(obj.ViewObject.ShapeColor,testbw=False)[1:]
                        outfile.write("usemtl color_" + mn + "\n")
                        materials.append(("color_" + mn,obj.ViewObject.ShapeColor,obj.ViewObject.Transparency))

            # write geometry
            for v in vlist:
                outfile.write("v" + v + "\n")
            for vn in vnlist:
                outfile.write("vn" + vn + "\n")
            for e in elist:
                outfile.write("l" + e + "\n")
            for f in flist:
                outfile.write("f" + f + "\n")

    outfile.close()
    FreeCAD.Console.PrintMessage(translate("Arch","Successfully written") + " " + filename + "\n")
    if materials:
        outfile = pythonopen(filenamemtl,"w")
        outfile.write("# FreeCAD v" + ver[0] + "." + ver[1] + " build" + ver[2] + " Arch module\n")
        outfile.write("# https://www.freecad.org\n")
        kinds = {"AmbientColor":"Ka ","DiffuseColor":"Kd ","SpecularColor":"Ks ","EmissiveColor":"Ke ","Transparency":"Tr ","Dissolve":"d "}
        done = [] # store names to avoid duplicates
        for mat in materials:
            if isinstance(mat,tuple):
                if not mat[0] in done:
                    outfile.write("newmtl " + mat[0] + "\n")
                    outfile.write("Kd " + str(mat[1][0]) + " " + str(mat[1][1]) + " " + str(mat[1][2]) + "\n")
                    outfile.write("Tr " + str(mat[2]/100) + "\n")
                    outfile.write("d " + str(1-mat[2]/100) + "\n")
                    done.append(mat[0])
            else:
                if not mat.Name in done:
                    outfile.write("newmtl " + mat.Name + "\n")
                    for prop in kinds:
                        if prop in mat.Material:
                            outfile.write(kinds[prop] + mat.Material[prop].strip("()").replace(',',' ') + "\n")
                    done.append(mat.Name)
        outfile.write("# Material Count: " + str(len(materials)))
        outfile.close()
        FreeCAD.Console.PrintMessage(translate("Arch","Successfully written") + ' ' + filenamemtl + "\n")


def open(filename):
    "called when freecad wants to open a file"
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname
    return insert(filename,doc.Name)

def insert(filename,docname):

    meshName = ntpath.basename(filename)
    for i in meshName.split():
        if "." in i:
            i = i.split(".")[0]
    meshName = i
    "called when freecad wants to import a file"
    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc

    with pythonopen(filename,"r") as infile:
        verts = []
        facets = []
        activeobject = None
        material = None
        colortable = {}
        content_array = []
        for line in infile:
            content_array.append(line)
    activeobjectExists = False
    for line in content_array:
        line = line.strip()
        if line[:2] == "o ":
            activeobjectExists = True
    if not activeobjectExists:
        activeobject = meshName
    for line in content_array:
        line = line.strip()
        if line[:7] == "mtllib ":
            matlib = os.path.join(os.path.dirname(filename),line[7:])
            if os.path.exists(matlib):
                with pythonopen(matlib,"r") as matfile:
                    mname = None
                    color = None
                    trans = None
                    for mline in matfile:
                        mline = mline.strip()
                        if mline[:7] == "newmtl ":
                            if mname and color:
                                colortable[mname] = [color,trans]
                            color = None
                            trans = None
                            mname = mline[7:]
                        elif mline[:3] == "Kd ":
                            color = tuple([float(i) for i in mline[3:].split()])
                        elif mline[:2] == "d ":
                            trans = int((1-float(mline[2:]))*100)
                    if mname and color:
                        colortable[mname] = [color,trans]
        elif line[:2] == "o ":
            if activeobject:
                makeMesh(doc,activeobject,verts,facets,material,colortable)
            material = None
            facets = []
            activeobject = line[2:]
        elif line[:2] == "v ":
            verts.append([float(i) for i in line[2:].split()])
        elif line[:2] == "f ":
            fa = []
            for i in line[2:].split():
                if "/" in i:
                    i = i.split("/")[0]
                fa.append(int(i))
            facets.append(fa)
        elif line[:7] == "usemtl ":
            material = line[7:]
    if activeobject:
        makeMesh(doc,activeobject,verts,facets,material,colortable)
    FreeCAD.Console.PrintMessage(translate("Arch","Successfully imported") + ' ' + filename + "\n")
    return doc

def makeMesh(doc,activeobject,verts,facets,material,colortable):
    mfacets = []
    if facets:
        for facet in facets:
            if len(facet) > 3:
                vecs = [FreeCAD.Vector(*verts[i-1]) for i in facet]
                vecs.append(vecs[0])
                pol = Part.makePolygon(vecs)
                try:
                    face = Part.Face(pol)
                except Part.OCCError:
                    print("Skipping non-planar polygon:",vecs)
                else:
                    tris = face.tessellate(1)
                    for tri in tris[1]:
                        mfacets.append([tris[0][i] for i in tri])
            else:
                mfacets.append([verts[i-1] for i in facet])
    if mfacets:
        mobj = doc.addObject("Mesh::Feature",activeobject)
        mobj.Label = activeobject
        mobj.Mesh = Mesh.Mesh(mfacets)
        if material and FreeCAD.GuiUp:
            if material in colortable:
                mobj.ViewObject.ShapeColor = colortable[material][0]
                if colortable[material][1] is not None:
                    mobj.ViewObject.Transparency = colortable[material][1]



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Metadata validation functions """

from typing import List

import FreeCAD

from Addon import Addon
from addonmanager_metadata import Metadata
import NetworkManager


class MetadataValidators:
    """A collection of tools for validating various pieces of metadata. Prints
    validation information to the console."""

    def validate_all(self, repos):
        """Developer tool: check all repos for validity and print report"""

        FreeCAD.Console.PrintMessage("\n\nADDON MANAGER DEVELOPER MODE CHECKS\n")
        FreeCAD.Console.PrintMessage("-----------------------------------\n")

        counter = 0
        for addon in repos:
            counter += 1
            if addon.metadata is not None:
                self.validate_package_xml(addon)
            elif addon.repo_type == Addon.Kind.MACRO:
                if addon.macro.parsed:
                    if len(addon.macro.icon) == 0 and len(addon.macro.xpm) == 0:
                        FreeCAD.Console.PrintMessage(
                            f"Macro '{addon.name}' does not have an icon\n"
                        )
            else:
                FreeCAD.Console.PrintMessage(
                    f"Addon '{addon.name}' does not have a package.xml file\n"
                )

        FreeCAD.Console.PrintMessage("-----------------------------------\n\n")

    def validate_package_xml(self, addon: Addon):
        """Check the package.xml file for the specified Addon"""
        if addon.metadata is None:
            return

        # The package.xml standard has some required elements that the basic XML
        # reader is not actually checking for. In developer mode, actually make sure
        # that all the rules are being followed for each element.

        errors = []

        errors.extend(self.validate_top_level(addon))
        errors.extend(self.validate_content(addon))

        if len(errors) > 0:
            FreeCAD.Console.PrintError(
                f"Errors found in package.xml file for '{addon.name}'\n"
            )
            for error in errors:
                FreeCAD.Console.PrintError(f"   * {error}\n")

    def validate_content(self, addon: Addon) -> List[str]:
        """Validate the Content items for this Addon"""
        errors = []
        contents = addon.metadata.content

        missing_icon = True
        if addon.metadata.icon and len(addon.metadata.icon) > 0:
            missing_icon = False
        else:
            if "workbench" in contents:
                wb = contents["workbench"][0]
                if wb.icon:
                    missing_icon = False
        if missing_icon:
            errors.append("No <icon> element found, or <icon> element is invalid")

        if "workbench" in contents:
            for wb in contents["workbench"]:
                errors.extend(self.validate_workbench_metadata(wb))

        if "preferencepack" in contents:
            for wb in contents["preferencepack"]:
                errors.extend(self.validate_preference_pack_metadata(wb))

        return errors

    def validate_top_level(self, addon: Addon) -> List[str]:
        """Check for the presence of the required top-level elements"""
        errors = []
        if not addon.metadata.name or len(addon.metadata.name) == 0:
            errors.append(
                "No top-level <name> element found, or <name> element is empty"
            )
        if not addon.metadata.version:
            errors.append(
                "No top-level <version> element found, or <version> element is invalid"
            )
        if not addon.metadata.description or len(addon.metadata.description) == 0:
            errors.append(
                "No top-level <description> element found, or <description> element "
                "is invalid"
            )

        maintainers = addon.metadata.maintainer
        if len(maintainers) == 0:
            errors.append("No top-level <maintainers> found, at least one is required")
        for maintainer in maintainers:
            if len(maintainer.email) == 0:
                errors.append(
                    f"No email address specified for maintainer '{maintainer.name}'"
                )

        licenses = addon.metadata.license
        if len(licenses) == 0:
            errors.append("No top-level <license> found, at least one is required")

        urls = addon.metadata.url
        errors.extend(self.validate_urls(urls))
        return errors

    @staticmethod
    def validate_urls(urls) -> List[str]:
        """Check the URLs provided by the addon"""
        errors = []
        if len(urls) == 0:
            errors.append(
                "No <url> elements found, at least a repo url must be provided"
            )
        else:
            found_repo = False
            found_readme = False
            for url in urls:
                if url["type"] == "repository":
                    found_repo = True
                    if len(url["branch"]) == 0:
                        errors.append(
                            "<repository> element is missing the 'branch' attribute"
                        )
                elif url["type"] == "readme":
                    found_readme = True
                    location = url["location"]
                    p = NetworkManager.AM_NETWORK_MANAGER.blocking_get(location)
                    if not p:
                        errors.append(
                            f"Could not access specified readme at {location}"
                        )
                    else:
                        p = p.data().decode("utf8")
                        if "<html" in p or "<!DOCTYPE html>" in p:
                            pass
                        else:
                            errors.append(
                                f"Readme data found at {location}"
                                " does not appear to be rendered HTML"
                            )
            if not found_repo:
                errors.append("No repo url specified")
            if not found_readme:
                errors.append(
                    "No readme url specified (not required, but highly recommended)"
                )
        return errors

    @staticmethod
    def validate_workbench_metadata(workbench: Metadata) -> List[str]:
        """Validate the required element(s) for a workbench"""
        errors = []
        if not workbench.classname or len(workbench.classname) == 0:
            errors.append("No <classname> specified for workbench")
        return errors

    @staticmethod
    def validate_preference_pack_metadata(pack: Metadata) -> List[str]:
        """Validate the required element(s) for a preference pack"""
        errors = []
        if not pack.name or len(pack.name) == 0:
            errors.append("No <name> specified for preference pack")
        return errors



#***************************************************************************
#*                                                                         *
#*   Copyright (c) 2022 Mario Passaglia                                    *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************


"""Provides the viewprovider code for the Shapestring object."""

import FreeCADGui as Gui

from draftviewproviders.view_base import ViewProviderDraft
from drafttaskpanels.task_shapestring import ShapeStringTaskPanelEdit

class ViewProviderShapeString(ViewProviderDraft):

    def __init__(self,vobj):

        vobj.Proxy = self

    def getIcon(self):

        return ":/icons/Draft_ShapeString.svg"

    def setEdit(self, vobj, mode):
        if mode != 0:
            return None

        if not "Draft_Edit" in Gui.listCommands(): # Using Draft_Edit to detect if the Draft, Arch or BIM WB has been loaded.
            self.wb_before_edit = Gui.activeWorkbench()
            Gui.activateWorkbench("DraftWorkbench")
        self.task = ShapeStringTaskPanelEdit(vobj)
        Gui.Control.showDialog(self.task)
        return True

    def unsetEdit(self, vobj, mode):
        if mode != 0:
            return None

        self.task.finish()
        if hasattr(self, "wb_before_edit"):
            Gui.activateWorkbench(self.wb_before_edit.name())
            delattr(self, "wb_before_edit")
        return True
